import random

from baopig import *


class UT_LayerPack_Zone(Zone):

    def __init__(self, *args, **kwargs):

        Zone.__init__(self, *args, **kwargs)

        classiclayer_tester = Zone(self, size=("100%", "50%"), name="classiclayer_tester")
        safe_layer = Layer(classiclayer_tester, Button, name="safe_layer")
        self.package = Zone(classiclayer_tester, padding=10, spacing=5, pos=(0, 35), size=(300, 300),
                            background_color=(90, 90, 80))
        Layer(self.package)

        def add():
            zone = Zone(self.package, name="zone", background_color="green4", padding=3, spacing=3,
                        pos=(random.randint(0, 200), random.randint(0, 200)))
            Rectangle(parent=zone, color=(255, 255, 0), size=(zone.rect.w, 10), pos=(0, 0), name="Cobaye")
            Text(zone, f"layer's length : {len(self.package.default_layer)}")
            Button(zone, "REMOVE", command=zone.kill)
            zone.default_layer.pack()
            zone.adapt(zone.default_layer)

        Button(classiclayer_tester, "ADD", command=add)
        b = Button(classiclayer_tester, "PACK", command=self.package.pack)
        # b.set_window((b.rect.left, b.rect.top, b.rect.w, 20),
        #              follow_movements=True)  # DONE : remove, here for some tests
        Indicator(b, "NOTE : By default, pack() sorts widgets by position", loc="bottom", max_width=200)
        Button(classiclayer_tester, "CLEAR", command=self.package.default_layer.clear)
        safe_layer.pack(axis="horizontal")
        Text(classiclayer_tester, pos=(330, 10), text="Testing Layer.pack()", font_height=30)

        s = Slider(classiclayer_tester, pos=(330, 80), step=1, minval=0, maxval=100, title="padding.left",
                   defaultval=self.package.padding.left, printed_title=True, length=150, wideness=15)

        def new_padding(val):
            self.package.padding._left = val

        s.signal.NEW_VAL.connect(new_padding, owner=None)

        s = Slider(classiclayer_tester, pos=(330, 120), step=1, minval=0, maxval=100, title="spacing.top",
                   defaultval=self.package.spacing.top, printed_title=True, length=150, wideness=15)

        def new_childrenmargins_top(val):
            self.package.spacing._top = val

        s.signal.NEW_VAL.connect(new_childrenmargins_top, owner=None)

        s = Slider(classiclayer_tester, pos=(330, 160), step=1, minval=0, maxval=100, title="spacing.bottom",
                   defaultval=self.package.spacing.bottom, printed_title=True, length=150, wideness=15)

        def new_childrenmargins_bottom(val):
            self.package.spacing._bottom = val

        s.signal.NEW_VAL.connect(new_childrenmargins_bottom, owner=None)
        Text(classiclayer_tester, "NOTE : Only spacing.top and spacing.left are being used",
             max_width=self.rect.w - s.rect.right - 10, ref=s, refloc="midright", midleft=(10, 0))

        # GRID TESTING
        gridlayer_tester = Zone(self, size=("100%", "50%"))
        self.default_layer.pack()
        safe_layer2 = Layer(gridlayer_tester, Button, name="safe_layer")
        self.package2 = Zone(gridlayer_tester, padding=10, spacing=5, pos=(0, 35), size=(300, 300),
                             background_color=(90, 90, 80))
        GridLayer(self.package2, nbrows=10, nbcols=10)

        def add():

            for i in range(10):

                if len(self.package2.children) == 81:
                    return

                searching = True
                x, y = 0, 0
                while searching:
                    x = random.randint(0, 8)
                    y = random.randint(0, 8)
                    if self.package2.default_layer.get_data(x, y) is None:
                        searching = False

                z = Zone(parent=self.package2, row=x, col=y, background_color="yellow4", size=(30, 30))
                Text(z, f"{x}, {y}", size=(30, 30))

        Button(gridlayer_tester, "ADD", command=add)
        Button(gridlayer_tester, "PACK", command=self.package2.pack)
        Button(gridlayer_tester, "CLEAR", command=self.package2.default_layer.clear)
        safe_layer2.pack(axis="horizontal")
        Text(gridlayer_tester, pos=(310, 10), text="Testing GridLayer.pack()", font_height=30)

        sliders_layer = Layer(gridlayer_tester, spacing=25)
        s = Slider(gridlayer_tester, layer=sliders_layer, step=1, minval=0, maxval=100, title="padding.left",
                   defaultval=self.package2.padding.left, printed_title=True, length=150, wideness=15)

        def new_paddingleft(val):
            self.package2.padding._left = val

        s.signal.NEW_VAL.connect(new_paddingleft, owner=None)

        s = Slider(gridlayer_tester, layer=sliders_layer, step=1, minval=0, maxval=100, title="padding.top",
                   defaultval=self.package2.padding.top, printed_title=True, length=150, wideness=15)

        def new_paddingtop(val):
            self.package2.padding._top = val

        s.signal.NEW_VAL.connect(new_paddingtop, owner=None)

        s = Slider(gridlayer_tester, layer=sliders_layer, step=1, minval=0, maxval=100, title="spacing.left",
                   defaultval=self.package2.spacing.left, printed_title=True, length=150, wideness=15)

        def new_childrenmargins_left(val):
            self.package2.spacing._left = val

        s.signal.NEW_VAL.connect(new_childrenmargins_left, owner=None)

        s = Slider(gridlayer_tester, layer=sliders_layer, step=1, minval=0, maxval=100, title="spacing.top",
                   defaultval=self.package2.spacing.top, printed_title=True, length=150, wideness=15)

        def new_childrenmargins_top(val):
            self.package2.spacing._top = val

        s.signal.NEW_VAL.connect(new_childrenmargins_top, owner=None)

        sliders_layer.pack(start_pos=(330, 80))


# For the PresentationScene import
ut_zone_class = UT_LayerPack_Zone

if __name__ == "__main__":
    from baopig.prefabs.testerscene import TesterScene
    app = Application()
    TesterScene(app, ut_zone_class)
    app.launch()




import inspect


class MemoryDict(dict):

    def __init__(self):

        dict.__init__(self)

        self._modified_keys = set()

    def __repr__(self):
        string = "{"
        for key in self.all_keys():
            string += f"'{key}': {self[key]}, "
        string += "}"
        return string

    def __setitem__(self, key, value):

        super().__setitem__(key, value)
        self._modified_keys.add(key)

    def all_keys(self):
        return set(self.keys())

    def subdict(self):
        return SubMemoryDict(self)


class SubMemoryDict(MemoryDict):

    def __init__(self, superdict):

        assert isinstance(superdict, MemoryDict)
        self._superdict = superdict
        MemoryDict.__init__(self)

    def __contains__(self, key):

        for key2 in self.keys():
            if key2 == key:
                return True
        return key in self._superdict

    def __getitem__(self, key):

        for key2, value in dict.items(self):
            if key2 == key:
                return value
        return self._superdict[key]

    def all_keys(self):
        return self._superdict.all_keys().union(set(self.keys()))


class StyleClass:

    def __init__(self):

        self._dict = MemoryDict()
        self._types = MemoryDict()
        self._constraints = MemoryDict()
        self._constraint_error_messages = MemoryDict()
        self._theme = None
        self._priority = 0
        self._priority_dict = MemoryDict()

    def __contains__(self, key):
        return key in self._dict

    def __getitem__(self, key):
        try:
            return self._dict[key]
        except KeyError:
            raise KeyError(f"'{key}' is not a style attribute (attributes={self._dict.all_keys()})")

    def __repr__(self):
        return f"Style{self._dict}"

    def _setitem(self, key, value):
        self._dict[key] = value
        self._priority_dict[key] = self._priority
        self.check_type(key)
        self.check_constraint(key)

    def _apply_on_instanciatedstyle(self, style):

        # SuperButton -> Button -> Container
        #      |            |
        #     red         blue            <- default style
        #                   |
        #                yellow           <- style for a whole application
        #
        # A SuperButton has to be red, there is a priority

        assert isinstance(style, InstanciatedStyle)
        for attr in "_dict", "_types", "_constraints", "_constraint_error_messages":
            defined_keys = getattr(style, attr).all_keys()
            for key, value in getattr(self, attr).items():
                if key in defined_keys:  # this attribute is already defined by a substyle
                    if style.get_priority(key) >= self.get_priority(key):
                        continue
                getattr(style, attr)[key] = value

    def check_constraint(self, key):
        """Check if the value at key follows the defined constraint"""

        constraint = self.get_constraint(key)
        if constraint is None:
            return
        value = self[key]
        try:
            ok = constraint(value)
            if not ok:
                raise ValueError(f"Wrong value for {key} : {value}")
        except Exception as e:
            custom_message = self.get_constraint_error_message(key)
            if custom_message is not None:
                raise e.__class__(str(e) + ", " + custom_message)
            raise e

    def check_type(self, key):
        """Check if the value at key matches with defined type, but don't apply the type"""

        required_type = self.get_type(key)
        if required_type is None:
            return
        value = self[key]
        if isinstance(value, required_type):
            return
        if isinstance(value, str) and value.startswith("theme-"):
            return
        try:
            _ = required_type(value)  # raise an error if this is impossible
        except Exception as e:  # TODO : TypeError
            raise e.__class__(str(e), f", value for {key} must be convertible in type {required_type}")

    def create(self, **kwargs):

        for key, value in kwargs.items():
            if key in self._dict:
                raise KeyError(f"This style attribute has already been created : {key}")
            self._setitem(key, value)

    def modify(self, **kwargs):
        """
        Works ike this :
        
        a_widget.style.modify(width=30)
        
        This will affect the appearance only if called before the Widget constructor
        """

        for key in self._dict.all_keys():
            if key in kwargs:
                self._setitem(key, kwargs.pop(key))

        if kwargs:
            raise KeyError(
                f"'{kwargs.keys()}' are not style attributes (attributes={self._dict.all_keys()})")

    def get_all_keys(self):
        """Return a set containing all style attributes"""
        return self._dict.all_keys()

    def get_constraint(self, key):

        try:
            return self._constraints[key]
        except KeyError:
            return None

    def get_constraint_error_message(self, key):

        try:
            return self._constraint_error_messages[key]
        except KeyError:
            return None

    def get_priority(self, key):

        return self._priority_dict[key]

    def get_type(self, key):

        try:
            return self._types[key]
        except KeyError:
            return None

    def get_modified_keys(self):
        """Return a generator containing all style attributes modified by this style, not its superstyle"""
        return self._dict.keys()

    def set_constraint(self, key, constraint, error_message=None):
        """
        bool(constraint(style[key])) must always return True
        If constraint is None, remove the current constraint
        WARNING : you can only set one constraint per key
        """

        if constraint is None:
            self._constraints[key] = lambda val: True  # prevent from getting super-style constraint
            if key in self._constraint_error_messages:
                self._constraint_error_messages.pop(key)
            return
        assert callable(constraint)
        assert len(inspect.signature(constraint).parameters) == 1, \
            "A constraint must only accept one parameter, the checked value"
        self._constraints[key] = constraint
        if error_message is not None:
            assert isinstance(error_message, str)
            self._constraint_error_messages[key] = error_message
        self.check_constraint(key)  # check constraint on current value

    def set_type(self, key, required_type):

        assert inspect.isclass(required_type)
        self._types[key] = required_type
        self.check_type(key)

    def substyle(self):

        return SubStyleClass(self)


class SubStyleClass(StyleClass):

    def __init__(self, superstyle):

        assert isinstance(superstyle, StyleClass)

        self._super = superstyle
        self._dict = superstyle._dict.subdict()
        self._types = superstyle._types.subdict()
        self._constraints = superstyle._constraints.subdict()
        self._constraint_error_messages = superstyle._constraint_error_messages.subdict()
        self._theme = None
        self._priority = superstyle._priority + 1
        self._priority_dict = superstyle._priority_dict.subdict()


class InstanciatedStyle(SubStyleClass):
    """InstanciatedStyle has copied values, types, constraints and error messages

    It is needed in order to stop every inheritance from the theme style"""
    def __init__(self, owner):

        self._owner = owner
        SubStyleClass.__init__(self, owner.__class__.STYLE.substyle())
        theme = owner.theme
        self._theme = theme

        # classes : all the super classes of the owner for wich a special style is defined in theme
        classes = (super_class for super_class in theme.get_set_classes()
                   if issubclass(owner.__class__, super_class))
        classes = sorted(classes, key=lambda cls: -len(cls.mro()))

        for super_class in classes:
            superclass_style = theme.get_style_for(super_class)
            superclass_style._apply_on_instanciatedstyle(self)

        for key in self._dict.all_keys():
            value = self[key]
            if hasattr(value, "copy") and not inspect.isclass(value):
                value = value.copy()
            if isinstance(value, str):
                if value.startswith("theme-"):
                    value = self._theme.get_value(value)
                elif value.startswith("CALCUL "):
                    value = eval(value[7:], {"self": owner})
            required_type = self.get_type(key)
            if required_type is not None:
                self._types[key] = required_type
                if not isinstance(value, required_type):
                    value = required_type(value)
            constraint = self.get_constraint(key)
            if constraint is not None:
                self._constraints[key] = constraint
            message = self.get_constraint_error_message(key)
            if message is not None:
                self._constraint_error_messages[key] = message
            self._dict[key] = value

    def _setitem(self, key, value):

        if isinstance(value, str) and value.startswith("theme-"):
            value = self._theme.get_value(value)
        required_type = self.get_type(key)
        if required_type is not None:
            if not isinstance(value, required_type):
                value = required_type(value)
        self._dict[key] = value
        self._priority_dict[key] = self._priority
        self.check_constraint(key)

    def set_theme(self, theme):

        if isinstance(theme, str):
            theme = get_subtheme_from_prefab(name=theme)
        assert isinstance(theme, Theme)
        self._theme = theme


class ThemeColors:

    def __init__(self, **kwargs):

        self.border = (0, 0, 0)
        self.content = (0, 213, 240)
        self.font = (0, 0, 0)
        self.font_opposite = (255, 255, 255)
        self.scene_background = (170, 170, 170)
        self.selection = (167, 213, 255)
        self.selection_rect = (107, 107, 205)

        for key in tuple(kwargs.keys()):
            if hasattr(self, key):
                setattr(self, key, kwargs.pop(key))
        if kwargs:
            raise KeyError(f"Wrong color keys : {kwargs.keys()}")

    def copy(self):

        return ThemeColors(**vars(self))


class Theme:
    """
    A theme is a dictionnary of style and other attributes

    If you change a theme's colors, it wont affect existing subthemes's colors
    """

    def __init__(self):
        self._all_styles = {}
        self.colors = ThemeColors()

    def __repr__(self):
        string = "Theme("
        classes = self.get_set_classes()
        for widget_class in classes:
            style = self.get_style_for(widget_class)
            string += f"\n\t{widget_class.__name__}: {style}"
        if classes:
            string += "\n"
        string += ")"
        return string

    def get_style_for(self, widget_class):
        """
        Return the defined style for a widget class in this theme.
        If there is no defined style, return the default style.
        
        WARNING : do never modify the returned style ! The consequences are tricky
                  use set_style_for() instead
        """

        for widget_class2, widget_defined_substyle in self._all_styles.items():
            if widget_class is widget_class2:
                return widget_defined_substyle
        return widget_class.STYLE

    def get_set_classes(self):
        """Return a set containing all widget classes whose style is defined by this theme"""
        return set(self._all_styles.keys())

    def get_value(self, string):
        assert isinstance(string, str) and string.startswith("theme-")
        value = string[6:]
        if value.startswith("color-"):
            return getattr(self.colors, value[6:])
        else:
            raise ValueError(f"Unknown attribute for theme : {string}")

    def issubtheme(self, theme):
        return False

    def set_style_for(self, widget_class, **style):

        if widget_class not in self._all_styles:
            self._all_styles[widget_class] = widget_class.STYLE.substyle()
        self._all_styles[widget_class].modify(**style)

    def subtheme(self):
        return SubTheme(self)


class SubTheme(Theme):

    def __init__(self, supertheme):
        assert isinstance(supertheme, Theme)
        self._super = supertheme
        self._all_styles = {}
        self.colors = supertheme.colors.copy()

    def __getitem__(self, key):

        return self.get_style_for(key)

    def get_style_for(self, widget_class):
        """
        WARNING : do never modify the returned style ! The consequences are tricky
                  use set_style_for() instead
        """

        for widget_class2, widget_defined_substyle in self._all_styles.items():
            if widget_class is widget_class2:
                return widget_defined_substyle
        return self._super.get_style_for(widget_class)

    def get_set_classes(self):
        """Return a set containing all widget classes whose style is defined by this theme"""
        return set(self._all_styles.keys()).union(self._super.get_set_classes())

    def issubtheme(self, theme):
        return theme is self._super

    def set_style_for(self, widget_class, **style):

        if widget_class not in self._all_styles:
            self._all_styles[widget_class] = self._super.get_style_for(widget_class).substyle()
        self._all_styles[widget_class].modify(**style)


class HasStyle:
    """
    A style is a dictionnary containing appearance attributes and values
    It is first inherited from parents then modified by the instance

    STYLE is a class attribute. It is the default style for the object.
    In order to properly override style from super class, you should always
    define style attributes as so:

    class MyWidget(SuperWidget):
        STYLE = SuperWidget.STYLE.substyle()
        STYLE["color"] = (0, 255, 0)

    A theme is a dictionnary of style and other attributes

    WARNING : you cannot change a widget's theme or style after its creation
    """

    STYLE = StyleClass()

    def __init__(self, arg, options=None):

        if hasattr(self, "_style"):  # previously called 'inherit_style()' for an anticipated style
            return

        if isinstance(arg, Theme):
            theme = arg
        elif isinstance(arg, str):
            theme = get_subtheme_from_prefab(name=arg)
        else:
            parent = arg
            self._parent = parent

            if "theme" in options:  # here, theme is the parent widget
                theme = options.pop("theme")
                if isinstance(theme, str):
                    theme = get_subtheme_from_prefab(name=theme)
                elif not theme.issubtheme(parent.theme):
                    raise PermissionError("Must be an parent sub-theme")
            else:
                theme = parent.theme.subtheme()

        self._theme = theme
        self._style = InstanciatedStyle(self)

        if options:
            style_attributes = {}
            for key, val in tuple(options.items()):
                if (val is not None) and (key in self.STYLE):
                    style_attributes[key] = options.pop(key)
            if style_attributes:
                self.style.modify(**style_attributes)

    style = property(lambda self: self._style)
    theme = property(lambda self: self._theme)

    def get_style_for(self, widget_class):
        """
        WARNING : do never modify the returned style ! The consequences are tricky
                  use set_style_for() instead
        """

        if not hasattr(self, "_theme"):
            raise PermissionError("You must inherit style before using it")
        return self.theme.get_style_for(widget_class)

    def set_style_for(self, *widget_classes, **kwargs):
        """
        Exemple :
        app.set_style_for(bp.Button, height=40)
        app.set_style_for(bp.Rectangle, color="blue", width=40)
        """

        if not hasattr(self, "_theme"):
            raise PermissionError("You must inherit style before using it")
        self.theme.set_style_for(*widget_classes, **kwargs)


def get_subtheme_from_prefab(name):
    from baopig.prefabs.themes import all_themes
    return all_themes[name].subtheme()





import pygame

from baopig.pybao.issomething import *
from .utilities import Color
from .widget import Widget


class Rectangle(Widget):
    """
    A Widget who is just a rectangle filled with one color

    If the border_width parameter is filled, the rectangle will not be filled, only its borders
    The border_width is given in pixels
    The border only goes inside the rect, not outside
    """

    STYLE = Widget.STYLE.substyle()
    STYLE.modify(
        width=30,
        height=30,
    )
    STYLE.create(
        color="theme-color-content",
        border_color="theme-color-border",
        border_width=0,
    )
    STYLE.set_type("color", Color)
    STYLE.set_type("border_color", Color)
    STYLE.set_type("border_width", int)
    STYLE.set_constraint("border_width", lambda val: val >= 0, "must be positive")

    def __init__(self, parent, **kwargs):

        Widget.__init__(self, parent, **kwargs)

        self._color = self.style["color"]
        self._border_color = self.style["border_color"]
        self._border_width = self.style["border_width"]

        self.send_paint_request()

    color = property(lambda self: self._color)
    border_color = property(lambda self: self._border_color)
    border_width = property(lambda self: self._border_width)

    def paint(self):
        self.surface.fill(self.color)
        if self.border_color is not None:
            pygame.draw.rect(self.surface, self.border_color, (0, 0) + self.rect.size, self.border_width * 2 - 1)
        # self.signal.NEW_SURFACE.emit()
        # self.send_display_request()

    def set_color(self, color=None):

        if color is None:
            color = (0, 0, 0, 0)
        self._color = Color(color)
        self.send_paint_request()

    def set_border_color(self, color):

        self._border_color = Color(color)
        self.send_paint_request()

    def set_border_width(self, border_width):

        assert 0 <= border_width
        self._border_width = border_width

        self.send_paint_request()


class Highlighter(Rectangle):
    """
    A Highlighter is a border filled with one color surrounding a target's rect
    If the highlighter can be in the target's layer, it is placed in front of the target
    The border is one pixel inside the rect, so targets like scenes can be visually
    highlighted
    """
    STYLE = Rectangle.STYLE.substyle()
    STYLE.modify(
        color=(0, 0, 0, 0),
        border_color="green",
        border_width=1,
    )

    def __init__(self, parent, target, **kwargs):

        if "ref" in kwargs:
            raise PermissionError("Use 'target' instead of 'ref'")
        if "size" in kwargs:
            raise PermissionError("A Highlighter's' size depends on its target")

        Rectangle.__init__(self, parent, ref=target, size=target.rect.size, **kwargs)

        self._target_ref = target.get_weakref()
        self.set_touchable_by_mouse(False)

        def handle_targetresize():
            self.resize(*self.target.rect.size)

        self.target.signal.RESIZE.connect(handle_targetresize, owner=self)

    target = property(lambda self: self._target_ref())


class Polygon(Widget):
    """
    Create a Polygon from vertices
    If offset is set, move all vertices by offset
    """

    def __init__(self, parent, color, vertices, width=0, offset=(0, 0), offset_angle=None, **kwargs):

        if True and 1:
            raise NotImplemented  # TODO

        assert "pos" not in kwargs, "Use offset instead"

        def plus(p1, p2):
            return p1[0] + p2[0], p1[1] + p2[1]

        def minus(p1, p2):
            return p1[0] - p2[0], p1[1] - p2[1]

        if offset_angle:
            import numpy
            rotation_matrix = numpy.array([[numpy.cos(offset_angle), -numpy.sin(offset_angle)],
                                           [numpy.sin(offset_angle), numpy.cos(offset_angle)]])
            vertices = tuple(rotation_matrix.dot(v) for v in vertices)

        topleft_corner = min(v[0] for v in vertices), min(v[1] for v in vertices)
        verts2 = tuple(minus(v, topleft_corner) for v in vertices)
        surf = pygame.Surface((
            max(v[0] for v in verts2)+1,
            max(v[1] for v in verts2)+1
        ), pygame.SRCALPHA)
        pygame.draw.polygon(surf, color, verts2, width)

        vertices = tuple(plus(offset, v) for v in vertices)
        pos = (
            min(v[0] for v in vertices),
            min(v[1] for v in vertices),
        )

        Widget.__init__(self, parent, surf, pos, **kwargs)

        self._vertices = vertices

    vertices = property(lambda self: self._vertices)


class Line(Widget):

    def __init__(self, parent, color, point_a, point_b, width=1, offset=(0, 0), offset_angle=None, **kwargs):

        assert "pos" not in kwargs, "Use offset instead"

        def plus(p1, p2):
            return p1[0] + p2[0], p1[1] + p2[1]

        def minus(p1, p2):
            return p1[0] - p2[0], p1[1] - p2[1]

        points = point_a, point_b
        if offset_angle:
            import numpy
            rotation_matrix = numpy.array([[numpy.cos(offset_angle), -numpy.sin(offset_angle)],
                                           [numpy.sin(offset_angle), numpy.cos(offset_angle)]])
            points = tuple(rotation_matrix.dot(p) for p in points)

        topleft_corner = min(p[0] for p in points), min(p[1] for p in points)
        points2 = tuple(minus(p, topleft_corner) for p in points)
        surf = pygame.Surface((
            max(p[0] for p in points2)+1,
            max(p[1] for p in points2)+1
        ), pygame.SRCALPHA)

        pygame.draw.line(surf, color, points2[0], points2[1], width)
        points = tuple(plus(offset, p) for p in points)
        pos = (
            min(p[0] for p in points),
            min(p[1] for p in points),
        )

        Widget.__init__(self, parent, surface=surf, pos=pos, **kwargs)

        self._points = points

    points = property(lambda self: self._points)


class Circle(Widget):

    def __init__(self, parent, color, center, radius, border_width=0, **kwargs):

        if isinstance(radius, float):
            radius = int(radius)
        assert "pos" not in kwargs, "Use center instead"
        assert isinstance(radius, int)
        if border_width > 1:
            raise NotImplemented
        surf = pygame.Surface((radius * 2, radius * 2), pygame.SRCALPHA)
        pygame.draw.circle(surf, color, (radius, radius), radius, border_width)
        Widget.__init__(self, parent, surface=surf, center=center, **kwargs)

        self._color = color
        self._radius = radius
        self._border_width = border_width

    color = property(lambda self: self._color)
    radius = property(lambda self: self._radius)
    border_width = property(lambda self: self._border_width)

    def set_radius(self, radius):

        if isinstance(radius, float):
            radius = int(radius)
        assert isinstance(radius, int)
        assert radius >= 0

        self._radius = radius
        surf = pygame.Surface((radius * 2, radius * 2), pygame.SRCALPHA)
        pygame.draw.circle(surf, self.color, (radius, radius), radius, self.border_width)
        self.set_surface(surf)





import pygame
from .utilities import paint_lock
from .layer import Layer


class RowOrCol:

    def __init__(self, grid, index):

        assert isinstance(grid, GridLayer)
        assert isinstance(index, int) and index >= 0

        self._grid = grid
        self._grid_layer = grid._layer
        self._grid_data = grid._data  # shared ressource
        self._index = index

    def __contains__(self, item):

        for widget in self.ichildren:
            if widget is item:
                return True
        return False

    def __len__(self):

        length = 0
        for _ in self.ichildren:
            length += 1
        return length

    children = property(lambda self: tuple(widget for widget in self if widget is not None))
    ichildren = property(lambda self: (widget for widget in self if widget is not None))
    is_first = property(lambda self: self._index == 0)

    def is_empty(self):

        for _ in self.ichildren:
            return False
        return True


class Column(RowOrCol):
    """
    A Column can be seen as a list of cells with the same 'col' value.
    It is the manager of GridLayer.cols and can read GridLayer.data
    """

    def __init__(self, grid, index):

        RowOrCol.__init__(self, grid, index)

        self._width = grid.col_width
        if self.is_first:
            self._left = grid.padding.left
        else:
            self._left = self.get_previous_col().right + grid.spacing.left

        grid._cols.append(self)

    def __iter__(self):

        for data_row in self._grid_data:
            yield data_row[self._index]

    def __getitem__(self, row):
        """
        Return the value of the cell in this col, indexed by 'col'
        """
        return self._grid_data[row][self._index]

    is_adaptable = property(lambda self: self._width is None)
    is_last = property(lambda self: self._index == len(self._grid.cols) - 1)
    left = property(lambda self: self._left)
    right = property(lambda self: self._left + self.get_width())

    def _update_width(self):
        """
        Adapt the col width to the required width
        Needed when, in an adaptative column, a widget is added, resized or removed
        NOTE : this is a heavy function, try to call it when you are sure it will change something
        """

        for widget in self.ichildren:
            self._grid._update_widget(widget)

        if not self.is_last:
            self.get_next_col()._update_left(self.left + self.get_width() + self._grid.spacing.left)

    def _update_left(self, left):

        dx = left - self.left
        if dx == 0:
            return
        self._left = left
        for widget in self.ichildren:
            self._grid._update_widget(widget)

        if not self.is_last:
            self.get_next_col()._update_left(left + self.get_width() + self._grid.spacing.left)

    def get_cell_rect(self, row):
        """
        Return the cell's position and size inside a tuple object
        """
        return self.left, self._grid.get_row(row).top, self.get_width(), self._grid.get_row(row).get_height()

    def get_next_col(self):

        if self.is_last:
            return
        return self._grid.get_col(self._index + 1)

    def get_previous_col(self):
        """
        Return the col above self
        """
        if self.is_first:
            return None
        return self._grid.get_col(self._index - 1)

    def get_width(self):

        if self._width is not None:
            return self._width

        if self.is_empty():
            return 0

        return max(widget.rect.w for widget in self if widget is not None)

    def set_width(self, width):
        """
        Set the col's width
        If width is None, the col will adapt to its widest widget
        """
        assert (width is None) or isinstance(width, (int, float)) and width >= 0
        if not self._grid.cols_are_adaptable:
            raise PermissionError("This grid has a fixed col_width : {}".format(self._grid.col_width))
        self._width = width
        self._update_width()


class Row(RowOrCol):
    """
    A Row can be seen as a list of cells with the same 'row' value.
    It is the manager of GridLayer.rows and can read GridLayer.data
    """
    def __init__(self, grid, index):

        RowOrCol.__init__(self, grid, index)

        self._height = grid.row_height
        if self.is_first:
            self._top = grid.padding.top
        else:
            self._top = self.get_previous_row().bottom + grid.spacing.top

        grid._rows.append(self)

    def __contains__(self, item):

        return self._grid_data[self._index].__contains__(item)

    def __getitem__(self, col):
        """
        Return the value of the cell in this row, indexed by 'col'
        """
        return self._grid_data[self._index][col]

    def __iter__(self):

        return self._grid_data[self._index].__iter__()

    bottom = property(lambda self: self.top + self.get_height())
    is_adaptable = property(lambda self: self._height is None)
    is_last = property(lambda self: self._index == len(self._grid.rows)-1)
    top = property(lambda self: self._top)

    def _update_height(self):
        """
        Adapt the row height to the required height
        """

        for widget in self.ichildren:
            self._grid._update_widget(widget)

        if not self.is_last:
            self.get_next_row()._update_top(self.top + self.get_height() + self._grid.spacing.top)

        # if not self.is_last:
        #     self.get_next_row()._update_top(top + h + self._grid.spacing.top)

    def _update_top(self, top):

        dy = top - self.top
        if dy == 0:
            return
        self._top = top
        for widget in self.ichildren:
            self._grid._update_widget(widget)

        if not self.is_last:
            self.get_next_row()._update_top(top + self.get_height() + self._grid.spacing.top)

    def get_cell_rect(self, col):
        """
        Return the cell's position and size inside a tuple object
        """
        return self._grid.get_col(col).left, self.top, self._grid.get_col(col).get_width(), self.get_height()

    def get_height(self):

        if self._height is not None:
            return self._height

        if self.is_empty():
            return 0

        return max(widget.rect.h for widget in self if widget is not None)

    def get_next_row(self):

        if self.is_last:
            return
        return self._grid.get_row(self._index+1)

    def get_previous_row(self):
        """
        Return the row above self
        """
        if self.is_first:
            return None
        return self._grid.get_row(self._index-1)

    def set_height(self, height):
        """
        Set the row's height
        If height is None, the row will adapt to its hightest widget
        """
        assert (height is None) or isinstance(height, (int, float)) and height >= 0
        if not self._grid.rows_are_adaptable:
            raise PermissionError("This grid has a fixed row_height : {}".format(self._grid.row_height))
        self._height = height
        self._update_height()


class GridLayer(Layer):
    """
    A GridLayer is a Layer who places its children itself, depending on their
    attributes 'row' and 'col'

    If nbrows is None, adding a widget will create missing rows if needed
    nbcols works the same

    A GridLayer dimension (row or column) have two implicit modes : adaptable and fixed
    If the dimension size (col_width or row_height) is set, the mode is fixed. Else,
    it is adaptable.
    Fixed means all cell have the same dimension size
    Adaptable means cells receive their dimension size from the largest cell in the
    dimension.
    Note that for one dimension, you can reset its size. It means yous can, in the
    same grid, have an adaptable row above a fixed row, as below :

    gl = GridLayer(some_parent, cols=None, rows=5, col_width=None, row_height=25)
    grid = gl.grid
                                       # All rows have the fixed mode
    grid.get_row(3).set_height(None)   # This row gets the adaptable mode
    grid.get_row(4).set_height(45)     # This row gets the fixed mode

    Two widgets can't fit in the same cell

    WARNING : manipulating a grid with multi-threading might cause issues
    """

    def __init__(self, *args, nbcols=None, nbrows=None, col_width=None, row_height=None, **kwargs):

        Layer.__init__(self, *args, **kwargs)

        if nbcols is not None:
            assert isinstance(nbcols, int) and nbcols > 0
        if nbrows is not None:
            assert isinstance(nbrows, int) and nbrows > 0

        self._layer = self
        self._nbcols = None
        self._nbrows = None
        self._col_width = col_width
        self._row_height = row_height

        self._data = [[None]]  # shared ressource initialized with one row and one column
        self._cols = []
        self._rows = []

        Row(self, 0)
        Column(self, 0)
        if nbcols:
            self.set_nbcols(nbcols)
        if nbrows:
            self.set_nbrows(nbrows)

    def __str__(self):
        return "{}(nbrows={}, nbcols={})".format(self.__class__.__name__, self.nbrows, self.nbcols)

    col_width = property(lambda self: self._col_width)
    cols = property(lambda self: self._cols)
    cols_are_adaptable = property(lambda self: self._col_width is None)
    nbcols = property(lambda self: self._nbcols if self._nbcols is not None else len(self._data[0]))
    nbrows = property(lambda self: self._nbrows if self._nbrows is not None else len(self._data))
    row_height = property(lambda self: self._row_height)
    rows = property(lambda self: self._rows)
    rows_are_adaptable = property(lambda self: self._row_height is None)

    def _update_widget(self, widget):
        """Updates window & position"""
        cell_rect = self.get_cell_rect(widget.row, widget.col)  # TODO : implement padding with cell_rect
        width = int(cell_rect[2] * widget._size_hints[0])
        height = int(cell_rect[3] * widget._size_hints[1])
        widget.resize(width, height)
        # TODO : use size_hint
        # widget.set_window_(cell_rect, follow_movements=False)
        widget.set_lock(pos=False)
        if widget.pos_manager.location is not None:
            widget.pos_manager.config(pos=getattr(pygame.Rect(cell_rect), widget.pos_manager.location),
                                      loc=widget.pos_manager.location)  # TODO : remove loc=
        else:
            widget.pos_manager.config(pos=cell_rect[:2])
        widget.set_lock(pos=True)

    def _update_size(self):

        with paint_lock:
            for row in self.rows:
                row._update_height()
            for col in self.cols:
                col._update_width()

    def accept(self, widget):
        """You must define at least the row or the column in order to insert a widget in a grid layer"""
        if (widget.col is None) or (widget.row is None):
            return False
        return super().accept(widget)

    def add(self, widget):

        if widget.pos_manager.reference is not self.container:
            raise PermissionError("Cannot use other ref than parent in a GridLayer")
        if widget.pos_manager.reference_location != "topleft":
            raise PermissionError("Cannot use other refloc than 'topleft' in a GridLayer")
        if (widget.col is None) or (widget.row is None):
            raise PermissionError(
                "You must define at least the row or the column in order to insert a widget in a GridLayer")
        try:
            super().add(widget)
            if self._nbcols is None and len(self.cols) - 1 < widget.col:
                self.set_nbcols(widget.col + 1)
                self._nbcols = None
            if self._nbrows is None and len(self.rows) - 1 < widget.row:
                self.set_nbrows(widget.row + 1)
                self._nbrows = None

            if self._data[widget.row][widget.col] is not None:
                raise PermissionError("Cannot insert {} at positon : row={}, col={}, because {} is already there"
                                      "".format(widget, widget.row, widget.col, self._data[widget.row][widget.col]))

            row = self.rows[widget.row]
            col = self.cols[widget.col]
            new_h = row.is_adaptable and widget.rect.height > row.get_height()
            new_w = col.is_adaptable and widget.rect.width > col.get_width()
            self._data[widget.row][widget.col] = widget
            if new_h:
                row._update_height()
            if new_w:
                col._update_width()

            if new_h or new_w or len(row) == 1 or len(col) == 1:
                self.pack()
            else:
                self._update_widget(widget)

            # don't need owner because, if the grid is killed,
            # it means the container is killed, so the widget is also killed
            widget.signal.RESIZE.connect(self._update_size, owner=self)
            widget.signal.KILL.connect(self._update_size, owner=self)

        except Exception as e:
            raise e

    def move(self, widget, col, row):  # TODO : invert col & row

        assert widget in self
        if self._data[row][col] is not None:
            if self._data[row][col] is widget:
                return
            raise PermissionError("This cell is already occupied by : " + str(self._data[row][col]))

        self._data[widget.row][widget.col] = None
        self._data[row][col] = widget
        widget._col = col
        widget._row = row
        if self.rows_are_adaptable or self.cols_are_adaptable:
            self.pack()
        else:
            self._update_widget(widget)

    def remove(self, widget):

        super().remove(widget)
        assert self._data[widget.row][widget.col] is widget

        col = self.get_col(widget.col)
        row = self.get_row(widget.row)
        old_col_width, old_row_height = None, None  # warning shut down
        if col.is_adaptable:
            old_col_width = col.get_width()
        if row.is_adaptable:
            old_row_height = row.get_height()

        widget.signal.RESIZE.disconnect(self._update_size)
        widget.signal.KILL.disconnect(self._update_size)
        self._data[widget.row][widget.col] = None

        if col.is_adaptable and (old_col_width != col.get_width()):
            col._update_width()
        if row.is_adaptable and (old_row_height != row.get_height()):
            row._update_height()

    def get_cell_rect(self, row, col):

        return self.get_col(col).get_cell_rect(row)

    def get_data(self, row, col):

        return self._data[row][col]

    def get_col(self, col_index):
        """
        Return the column from index
        """
        return self._cols[col_index]

    def get_row(self, row_index):
        """
        Return the row from index
        """
        return self._rows[row_index]

    def pack(self, start_pos=(0, 0), **kwargs):
        """
        Updates from spacing, not col.width & row.height
        """
        if kwargs:
            raise PermissionError("GridLayer.pack() only supports 'start_pos' parameters")

        with paint_lock:

            for col in self._cols:
                if col.is_first:
                    col._left = self.padding.left + start_pos[0]
                else:
                    col._left = col.get_previous_col().right + self.spacing.left

            for row in self._rows:
                if row.is_first:
                    row._top = self.padding.top + start_pos[1]
                else:
                    row._top = row.get_previous_row().bottom + self.spacing.top

                for widget in row.ichildren:
                    self._update_widget(widget)

    def set_col_width(self, width):

        assert (width is None) or isinstance(width, (int, float)) and width >= 0

        self._col_width = None
        for col in self.cols:
            col.set_width(width)
        self._col_width = width

    def set_nbcols(self, nbcols):
        """
        Set the number of columns
        If nbcols is None, adding a widget will create missing columns if needed
        """
        assert isinstance(nbcols, int) and nbcols > 0, nbcols
        if nbcols is not None:
            nbnew = nbcols - self.nbcols
            if nbnew < 0:
                raise PermissionError("Cannot reduce nbcols")
            elif nbnew > 0:
                # Create new columns
                for row in self._data:
                    row += [None] * nbnew
                for i in range(nbnew):
                    Column(self, i + nbcols - nbnew)
        self._nbcols = nbcols

    def set_nbrows(self, nbrows):
        """
        Set the number of rows
        If nbrows is None, adding a widget will create missing rows if needed
        """
        assert isinstance(nbrows, int) and nbrows > 0
        if nbrows is not None:
            nbnew = nbrows - self.nbrows
            if nbrows < 0:
                raise PermissionError("Cannot reduce nbrows")
            elif nbnew > 0:
                # Create new rows
                for i in range(self.nbrows, nbrows):
                    self._data += [[None] * self.nbcols]
                    Row(self, i)
        self._nbrows = nbrows

    def set_row_height(self, height):

        assert (height is None) or isinstance(height, (int, float)) and height >= 0

        self._row_height = None
        for row in self.rows:
            row.set_height(height)
        self._row_height = height

    def swap(self, widget1, widget2):

        assert widget1 in self
        assert widget2 in self

        widget1.set_lock(pos=False)
        widget2.set_lock(pos=False)

        self._data[widget1.row][widget1.col] = widget2
        self._data[widget2.row][widget2.col] = widget1

        widget1._col, widget1._row, widget2._col, widget2._row = widget2._col, widget2._row, widget1._col, widget1._row

        if self.rows_are_adaptable or self.cols_are_adaptable:
            self.pack()
        else:
            self._update_widget(widget1)
            self._update_widget(widget2)



import pygame
from baopig.io import LOGGER
from baopig.lib import Validable
from .lineedit import LineEdit


class Entry(LineEdit, Validable):
    """
    An Entry is a LineEdit who can be validated
    """

    def __init__(self, parent, entry_type, command=None, validate_on_defocus=True, **kwargs):

        """

        :param parent:
        :param entry_type: str, int, float...
        :param command: function executed on validation, with self.text as parameter
        :param args:
        :param validate_on_defocus:
        :param kwargs:
        """

        LineEdit.__init__(self, parent, **kwargs)
        Validable.__init__(self, catching_errors=False)

        self._entry_type = entry_type
        self.command = command
        self.validate_on_defocus = validate_on_defocus

        # if validate_on_defocus:
        #     self.signal.DEFOCUS_.connect(self.validate, owner=self)

    def accept(self, text):

        try:
            self._entry_type(text)
        except ValueError as e:
            LOGGER.warning(e)
            return False
        return True

    def handle_defocus(self):

        super().handle_defocus()
        if self.validate_on_defocus:
            self.validate()

    def handle_keydown(self, key):

        if key == pygame.K_RETURN:
            if self.accept(self.text):
                self.validate()
        else:
            super().handle_keydown(key)

    def handle_validate(self):

        if self.command is not None:
            self.command(self.text)
        self.defocus()
        if self.text:
            return self._entry_type(self.text)




import math


MILLISECONDS = .001
SECONDS = 1
MINUTES = 60
HOURS = 60 * MINUTES
DAYS = 24 * HOURS


def format_time(time, formatter=None):
    """
    Function who create a string from a time and a formatter
    The time value is considered as seconds

    It uses letters in formatter in order to create the appropriate string:
        - %D : days
        - %H : hours
        - %M : minutes
        - %S : seconds
        - %m : milliseconds

    Examples :

        format_time(time=600, foramtter="%H hours and %M minutes") -> "10 hours and 0 minutes"

    """

    formatter_is_unset = formatter is None
    if formatter is None:
        formatter = "%D:%H:%M:%S:%m"

    require_days = formatter.count("%D")
    require_hours = formatter.count("%H")
    require_minutes = formatter.count("%M")
    require_seconds = formatter.count("%S")
    require_milliseconds = formatter.count("%m")

    # Days
    days = int(time // 86400)
    if require_days:
        formatter = formatter.replace("%D", "{:0>2}".format(days))
        time = time % 86400

    # Hours
    hours = int(time // 3600)
    if require_hours:
        formatter = formatter.replace("%H", "{:0>2}".format(hours))
        time = time % 3600

    # Minutes
    minutes = int(time // 60)
    if require_minutes:
        formatter = formatter.replace("%M", "{:0>2}".format(minutes))
        time = time % 60

    # Seconds
    if require_seconds:
        formatter = formatter.replace("%S", "{:0>2}".format(int(time)))
        time = time % 1

    # milliseconds
    if require_milliseconds:
        formatter = formatter.replace("%m", "{:0>3}".format(int(time * 1000)))

    if formatter_is_unset:
        while formatter.startswith("00:") and len(formatter) > 6:
            formatter = formatter[3:]

    return formatter


def present_time(time, max_unit=SECONDS, min_unit=MILLISECONDS):
    """
    Arrondi au superieur
    If min_unit is set, the time will be given in the unit
    If max_unit is set, the time will contain every unit from min_unit to max_unit
    max_unit cannot be lower than SECONDS
    """
    # Round the time
    # If time = 1.5 seconds and min_unit is SECONDS, then we round it to 2
    time = math.ceil(time / min_unit) * min_unit

    res = ""
    got_max_unit = False

    # Hours
    hours = int(time // HOURS)
    if min_unit >= HOURS:
        return res + str(hours)
    if hours or max_unit >= HOURS:
        got_max_unit = True
        res += str(hours) + ":"
        time = time % HOURS

    # Minutes
    minutes = int(time // MINUTES)
    if min_unit >= MINUTES:
        return res + ("{:0>2}".format(minutes) if got_max_unit else str(minutes))
    if got_max_unit:
        res += "{:0>2}:".format(minutes)
        if minutes:
            time = time % MINUTES
    elif minutes or max_unit >= MINUTES:
        got_max_unit = True
        res += str(minutes) + ":"
        time = time % MINUTES

    # Seconds
    seconds = int(time // SECONDS)
    if min_unit >= SECONDS:
        return res + ("{:0>2}".format(seconds) if got_max_unit else str(seconds))
    if got_max_unit:
        res += "{:0>2}:".format(seconds)
    else:  # If the time is .345, we want 0:345
        res += str(seconds) + ":"
    if seconds:
        time = time % SECONDS

    # milliseconds
    milliseconds = int(time // MILLISECONDS)
    assert min_unit >= MILLISECONDS, "Wrong min_unit value"
    return res + "{:0>3}".format(milliseconds)


# TESTS UNITAIRES
"""
time = 400 * 60 + .8
print("Time :", time)
print(format_time(time, "Days : %D"))
print(format_time(time, "Hours : %H"))
print(format_time(time, "Minutes : %M"))
print(format_time(time, "Seconds : %S"))
print(format_time(time, "Milliseconds : %m"))

print(format_time(time, "%D:%H:%M:%S:%m"))
print(format_time(time, "%H:%M:%S:%m"))
print(format_time(time, "%M:%S:%m"))
print(format_time(time))

print(present_time(time, min_unit=MINUTES))
"""




from .entry import Entry, LOGGER


class NumEntry(Entry):
    """
    A NumEntry is an Entry who only accepts numbers
    It can accept a defined range of numbers
    """
    # TODO : work on it, right now it's awful
    # TODO : Writing a long number va faire une retour à la ligne
    STYLE = Entry.STYLE.substyle()
    STYLE.modify(
        width=40
    )

    def __init__(self, parent, minval=None, maxval=None, default=None,
                 accept_floats=True, *args, **kwargs):

        entry_type = float if accept_floats else int
        Entry.__init__(self, parent, entry_type=entry_type, *args, **kwargs)

        self._minval = minval
        self._maxval = maxval
        self._accepted_numbers = None  # TODO : use or remove ?
        if default is not None:
            assert self.accept(default)
            self.set_text(str(default))

    def accept(self, text):

        try:
            value = self._entry_type(text)
        except ValueError as e:
            LOGGER.warning(e)
            return False

        if self._minval and value < self._minval:
            LOGGER.warning("Wrong value : {}, the minimum value is {}".format(value, self._minval))
            return False

        if self._maxval and value > self._maxval:
            LOGGER.warning("Wrong value : {}, the maximum value is {}".format(value, self._maxval))
            return False

        if self._accepted_numbers and value not in self._accepted_numbers:
            LOGGER.warning("Wrong value : {}, must be one of {}".format(value, self._accepted_numbers))
            return False

        return True




# TODO : smart to use static, interactive, dynamic ?

import pygame
from baopig.io import mouse
from baopig.documentation import Widget as WidgetDoc
from baopig.communicative import Communicative
from .utilities import paint_lock, MetaPaintLocker
from .style import HasStyle


class WeakRef:
    def __init__(self, ref):
        self._ref = ref

    def __call__(self):
        return self._ref


class _PosManager:
    """
    An Origin is referenced by its parent
    When the parent moves, the widget follows

    An Origin coordinate can be given in different ways:
        - x = 4         sets x to 4 pixels
        - x = '10%'     sets x to self.parent.width * 10 / 100 (automatically updated)

    WARNING : A widget with a dynamic position (pourcentage position) cannot be manually
              moved by any other way than redefining the WTF ?? TODO
    """

    def __init__(self, owner):

        # asked_pos is the distance between reference at reference_location and owner at location
        self._asked_pos = owner.style["pos"]
        self._location = owner.style["loc"]
        reference = owner.style["ref"]
        self._reference_location = owner.style["refloc"]
        self._referenced_by_hitbox = owner.style["referenced_by_hitbox"]

        if reference is None:
            reference = owner.parent
        self._owner_ref = owner.get_weakref()
        self._reference_ref = reference.get_weakref()

    def __str__(self):
        return f"Origin(asked_pos={self.asked_pos}, location={self.location}, reference={self.reference}, " \
               f"reference_location={self.reference_location}, referenced_by_hitbox={self.referenced_by_hitbox}, " \
               f"is_locked={self.is_locked})"

    asked_pos = property(lambda self: self._asked_pos)
    referenced_by_hitbox = property(lambda self: self._referenced_by_hitbox)
    is_locked = property(lambda self: self.owner.has_locked("pos"))
    location = property(lambda self: self._location)
    owner = property(lambda self: self._owner_ref())
    reference = property(lambda self: self._reference_ref())
    reference_location = property(lambda self: self._reference_location)

    def _get_sticky(self):
        if self.location == self.reference_location:
            return self.location
        return None

    sticky = property(_get_sticky)

    def _reset_asked_pos(self):
        """
        This method should only be called by Widget._move()
        If the asked_pos was including percentage, and the percentage don't match anymore, it will be
        replaced by an integer value
        """

        old_pos = self.get_pos_relative_to_owner_parent()
        current_pos = getattr(self.owner.rect, self.location)
        if old_pos == current_pos:
            return

        if self.is_locked:
            raise AssertionError(self)

        owner_abspos_at_location = getattr(self.owner.abs_rect, self.location)
        reference_abspos_at_location = getattr(self.reference.abs_rect, self.reference_location)

        self._asked_pos = (
            owner_abspos_at_location[0] - reference_abspos_at_location[0],
            owner_abspos_at_location[1] - reference_abspos_at_location[1]
        )

    @staticmethod
    def accept(coord):

        if isinstance(coord, str):
            if not coord.endswith('%'):
                return False
            try:
                int(coord[:-1])
                return True
            except ValueError:
                return False

        elif hasattr(coord, "__iter__"):
            if len(coord) != 2:
                return False
            return False not in (_PosManager.accept(c) for c in coord)

        try:
            coord = int(coord)
            return True
        except ValueError:
            raise TypeError("Wrong position value : {} (see documentation above)".format(coord))

    def config(self, pos=None, loc=None, refloc=None, referenced_by_hitbox=None, sticky=None, locked=None):

        if locked is False:
            self.owner.set_lock(pos=False)

        if self.is_locked:
            raise PermissionError("This widget's position is locked")

        if sticky is not None:
            assert loc is None
            assert refloc is None
            loc = refloc = sticky

        if loc is not None:
            self._location = Location(loc)

        if refloc is not None:
            self._reference_location = Location(refloc)

        if referenced_by_hitbox is not None:
            self._referenced_by_hitbox = bool(referenced_by_hitbox)

        if pos is not None:
            assert _PosManager.accept(pos), f"Wrong position value : {pos} (see documentation above)"
            self._asked_pos = pos

        self.owner._update_pos()

        if locked is True:
            self.owner.set_lock(pos=True)

    def get_pos_relative_to_owner_parent(self):

        pos = []

        rect = "hitbox" if self.referenced_by_hitbox else "rect"

        # Percentage translation of asked_pos
        ref_rect = getattr(self.reference, rect)
        for i, c in enumerate(self._asked_pos):
            if isinstance(c, str):
                c = ref_rect.size[i] * float(c[:-1]) / 100
            pos.append(int(c))

        # Transition at reference_location
        if self._reference_location != "topleft":
            ref_autorect = getattr(self.reference, "auto_" + rect)
            d = getattr(ref_autorect, self._reference_location)
            for i in range(2):
                pos[i] += d[i]

        # Set pos relative to self.owner.parent
        if self.reference is not self.owner.parent:
            ref_absrect = getattr(self.reference, "abs_" + rect)
            pos = (
                pos[0] + ref_absrect.left - self.owner.parent.abs_rect.left,
                pos[1] + ref_absrect.top - self.owner.parent.abs_rect.top
            )

        return tuple(pos)

    pos = property(get_pos_relative_to_owner_parent)


class _Window:  # TODO : Layouts -> no more window & no more hitbox

    def __init__(self, owner):

        self._owner = owner

        self._is_following_movements = True
        self._is_set = False
        self._offset = (0, 0)  # referenced from owner
        self._size = owner.rect.size
        self._surface = None

        # self._update_surface_TBR()

    is_following_movements = property(lambda self: self._is_following_movements)
    is_set = property(lambda self: self._is_set)
    offset = property(lambda self: self._offset)
    size = property(lambda self: self._size)
    surface = property(lambda self: self._surface)

    def _update_surface_TBR(self):

        subsurface_rect = self._owner.auto_rect.clip(self._offset + self.size)
        self._surface = self._owner.surface.subsurface(subsurface_rect)

    def config(self, offset=None, size=None):

        if offset is not None:
            self._offset = tuple(offset)

        if size is not None:
            self._size = tuple(size)

        # self._update_surface_TBR()

    def get_hitbox(self):

        offset_refed_from_owner_parent = (
            self._offset[0] + self._owner.rect.left,
            self._offset[1] + self._owner.rect.top,
        )

        return pygame.Rect(offset_refed_from_owner_parent, self._size).clip(self._owner.rect)


class Location(str):
    ACCEPTED = (
        "topleft", "midtop", "topright",
        "midleft", "center", "midright",
        "bottomleft", "midbottom", "bottomright",
    )

    def __new__(cls, location):
        if location not in Location.ACCEPTED:
            raise ValueError(f"Wrong value for location : '{location}', "
                             f"must be one of {Location.ACCEPTED}")

        return str.__new__(cls, location)


class ProtectedHitbox(pygame.Rect):
    """
    A ProtectedHitbox cannot be resized or moved
    """
    ERR = PermissionError("A ProtectedHitbox cannot change at all")

    def __init__(self, *args, **kwargs):
        pygame.Rect.__init__(self, *args, **kwargs)

    def __setattr__(self, *args):
        raise self.ERR

    pos = property(lambda self: tuple(self[:2]))

    def clamp_ip(self, rect):
        raise self.ERR

    def inflate_ip(self, x, y):
        raise self.ERR

    def move_ip(self, x, y):
        raise self.ERR

    def referencing(self, pos):
        """
        Return a new pos, relative the hitbox topleft
        The pos is considered as relative to (0, 0)

        Example :
            hibtox = ProtectedHitbox(100, 200, 30, 30)
            pos = (105, 205)
            hitbox.referencing(pos) -> (5, 5)
        """
        return pos[0] - self.left, pos[1] - self.top

    def union_ip(self, rect):
        raise self.ERR

    def unionall_ip(self, rect_sequence):
        raise self.ERR


class WidgetCore(WidgetDoc, Communicative):

    def __init__(self, parent):
        assert hasattr(parent, "_warn_change")

        Communicative.__init__(self)

        # NOTE : Since self is a parent's child, it doesn't need to use weakrefs
        self._parent = parent
        self.__scene = parent.scene
        self._application = parent.application

        # weakref will return None after widget.kill()
        self._weakref = WeakRef(self)
        self.create_signal("KILL")

    application = property(lambda self: self._application)
    parent = property(lambda self: self._parent)
    scene = property(lambda self: self.__scene)

    is_alive = property(lambda self: self._weakref() is not None)
    is_dead = property(lambda self: self._weakref() is None)

    def get_weakref(self, callback=None):
        """
        A weakref is a reference to an object
        callback is a function called when the widget die

        Example :
            weak_ref = my_widget.get_weakref()
            my_widget2 = weak_ref()
            my_widget is my_widget2 -> return True
            del my_widget
            print(my_widget2) -> return None
        """
        if callback is not None:
            assert callable(callback)
            self.signal.KILL.connect(callback, owner=None)
        return self._weakref

    def kill(self):

        if not self.is_alive:
            return

        with paint_lock:
            if self.parent is not None:  # False when called by Widget.wake()
                self.parent._remove_child(self)
            self.signal.KILL.emit(self._weakref)
            self.disconnect()
            self._weakref._ref = None

        del self


class HasLock:  # TODO : remove

    def __init__(self):

        class Locks:
            pos = False
            width = False
            height = False
            size = False
            visibility = False

        self._has_locked = Locks()

    def has_locked(self, key):

        try:
            return getattr(self._has_locked, key)
        except AttributeError:
            raise KeyError(f"Unknown key: {key}, availible keys are:{tuple(self._has_locked.__dict__.keys())}")

    def set_lock(self, **kwargs):
        """
        In kwargs, keys can be:
            pos
            width
            height
            size
            visibility
        In kwargs, values are interpreted as booleans
        """
        for key, locked in kwargs.items():
            if not hasattr(self._has_locked, key):
                raise KeyError(f"Unknown key: {key}, availible keys are:{tuple(self._has_locked.__dict__.keys())}")

            self._has_locked.__setattr__(key, bool(locked))
            if key in ("height", "width"):
                self._has_locked.size = self._has_locked.height and self._has_locked.width
            elif key == "size":
                self._has_locked.height = bool(locked)
                self._has_locked.width = bool(locked)


class Widget_VisibleSleepy(WidgetCore, HasLock):

    def __init__(self, parent):

        WidgetCore.__init__(self, parent)
        HasLock.__init__(self)

        # Sleep
        self._is_asleep = False
        self._sleep_parent_ref = lambda: ...
        self.create_signal("SLEEP")
        self.create_signal("WAKE")

        # Visibility
        self._is_visible = True
        self.create_signal("SHOW")
        self.create_signal("HIDE")

    is_asleep = property(lambda self: self._is_asleep)
    is_awake = property(lambda self: not self._is_asleep)
    is_hidden = property(lambda self: not self._is_visible)
    is_visible = property(lambda self: self._is_visible)

    def hide(self):

        if self._has_locked.visibility:
            return

        if not self.is_visible:
            return

        self._is_visible = False

        self.send_display_request()
        self.signal.HIDE.emit()

    def show(self):

        if self._has_locked.visibility:
            return
        if self.is_visible:
            return

        self._is_visible = True
        self.send_display_request()
        self.signal.SHOW.emit()

    def sleep(self):

        if self.is_asleep:
            return

        if self in self.parent.children:  # False when called during construction
            self.parent._remove_child(self)

        self._sleep_parent_ref = self.parent.get_weakref()
        self._parent = None
        self._is_asleep = True
        self.signal.SLEEP.emit()

    def wake(self):

        if self.is_awake:
            return

        self._parent = self._sleep_parent_ref()
        self._sleep_parent_ref = lambda: None
        if self.parent is None:
            return self.kill()

        self._is_asleep = False
        self.parent._add_child(self)
        self.signal.WAKE.emit()


class TouchableByMouse:

    def __init__(self, touchable):
        self._is_touchable_by_mouse = touchable

    is_touchable_by_mouse = property(lambda self: self._is_touchable_by_mouse)

    def set_touchable_by_mouse(self, val):
        self._is_touchable_by_mouse = bool(val)


class HasProtectedHitbox(Widget_VisibleSleepy, HasStyle, TouchableByMouse):
    STYLE = HasStyle.STYLE
    STYLE.create(
        pos=(0, 0),
        loc="topleft",
        ref=None,  # default is parent
        refloc="topleft",
        referenced_by_hitbox=False,
        width=None,  # must be filled
        height=None,  # must be filled
    )
    STYLE.set_type("loc", Location)
    STYLE.set_type("refloc", Location)

    def __init__(self, parent, size, touchable=True, **kwargs):
        """
        rect is the surface hitbox, relative to the parent
        abs_rect is the rect relative to the application
        auto_rect is the rect relative to the widget itself -> topleft = (0, 0)
        window is the rect inside wich the surface can be seen, relative to the parent
        hitbox is the result of clipping rect and window.
        If window is set to None, the hitbox will equal to the rect
        abs_hitbox is the hitbox relative to the application
        auto_hitbox is the hitbox relative to the widget itself
        """

        Widget_VisibleSleepy.__init__(self, parent)
        HasStyle.__init__(self, parent, options=kwargs)
        TouchableByMouse.__init__(self, touchable)

        # Shortcuts :                                             - NOTE : Can only use one shortcut
        #   sticky="center"   <=>  loc="center", refloc="center"
        #   center=(200, 45)  <=>  pos=(200, 45), loc="center"    - works for every location
        if "sticky" in kwargs:
            sticky = Location(kwargs.pop("sticky"))
            self.style.modify(loc=sticky, refloc=sticky)
        else:
            for key in tuple(kwargs.keys()):
                if key in Location.ACCEPTED:
                    self.style.modify(pos=kwargs.pop(key), loc=key)
                    break

        if kwargs:
            raise ValueError(f"Unused options : {kwargs}")

        """
        MOTION is emitted when the absolute position of the widget.rect moves
        It sends two parameters : dx and dy
        WARNING : the dx and dy are the motion of the widget.rect ! This means, when the
                  widget's parent moves, a MOTION.emit(dx=0, dy=0) is probalby raised.
                  In facts, if the widget.pos_manager.reference is not the parent, the motion
                  won't be dx=0, dy=0
        NOTE : an hitbox cannot move and be resized in the same time
        """
        self.create_signal("MOTION")

        """
        Change a surface via set_surface(...) is the only way to change a cmponent size
        Changing size will emit self.signal.RESIZE if the widget is visible
        A widget with locked size cannot be resized, but the surface can change
        
        When a widget is resized, we need a point of reference (called location) whose pixel
        will not move
        
        pos_manager.location can be one of : topleft,      midtop,     topright,
                                             midleft,      center,     midright,
                                             bottomleft,   midbottom,  bottomright
        
        Example : pos_manager.location = midright
                  if the new size is 10 more pixel on the width, then the position is on the right,  TODO
                  so we add the 10 new pixels to the left 
        """
        self.create_signal("RESIZE")

        # This will initialize the rects and hiboxes
        self._pos_manager = _PosManager(owner=self)

        self._rect = ProtectedHitbox((0, 0), size)
        self._abs_rect = ProtectedHitbox((0, 0), size)
        self._auto_rect = ProtectedHitbox((0, 0), size)
        self._window = _Window(self)
        self._hitbox = ProtectedHitbox((0, 0), size)
        self._abs_hitbox = ProtectedHitbox((0, 0), size)
        self._auto_hitbox = ProtectedHitbox((0, 0), size)

        # SETUP
        pygame.Rect.__setattr__(self.rect, self._pos_manager.location, self._pos_manager.pos)
        pygame.Rect.__setattr__(self.abs_rect, "topleft",
                                (self.parent.abs_rect.left + self.rect.left, self.parent.abs_rect.top + self.rect.top))
        pygame.Rect.__setattr__(self.hitbox, "topleft", self.rect.topleft)
        pygame.Rect.__setattr__(self.abs_hitbox, "topleft", self.abs_rect.topleft)

        # Connections
        self.signal.WAKE.connect(self._update_pos, owner=self)
        pos_ref = self._pos_manager.reference
        pos_ref.signal.RESIZE.connect(self._update_pos, owner=self)
        if pos_ref != self.parent:
            pos_ref.signal.MOTION.connect(self._update_pos, owner=self)

        def update_pos_from_parent_movement():
            # Code from _update_pos(), without these two lines:
            #     if new_pos == old_pos:
            #       return
            # This allows the MOTION.emit() even if no movement has been made
            # This way, widgets referenced by this widget can update their positions
            if self.is_asleep:  # the widget has no parent
                return
            new_pos = self._pos_manager.get_pos_relative_to_owner_parent()
            rect = self.hitbox if self._pos_manager.referenced_by_hitbox else self.rect
            old_pos = getattr(rect, self._pos_manager.location)
            self._move(dx=new_pos[0] - old_pos[0], dy=new_pos[1] - old_pos[1])

        self.parent.signal.MOTION.connect(update_pos_from_parent_movement, owner=self)

    # ORIGIN
    pos_manager = property(lambda self: self._pos_manager)

    # HITBOX
    rect = property(lambda self: self._rect)
    abs_rect = property(lambda self: self._abs_rect)
    auto_rect = property(lambda self: self._auto_rect)
    window = property(lambda self: self._window)
    hitbox = property(lambda self: self._hitbox)
    abs_hitbox = property(lambda self: self._abs_hitbox)
    auto_hitbox = property(lambda self: self._auto_hitbox)

    def _move(self, dx, dy):

        if self.is_asleep:
            raise PermissionError("Asleep widgets cannot move")

        old_hitbox = tuple(self.hitbox)
        with paint_lock:

            pygame.Rect.__setattr__(self.rect, "left", self.rect.left + dx)
            pygame.Rect.__setattr__(self.rect, "top", self.rect.top + dy)
            pygame.Rect.__setattr__(self.abs_rect, "topleft", (self.parent.abs_rect.left + self.rect.left,
                                                               self.parent.abs_rect.top + self.rect.top))

            if self.window.is_set:

                if self.window.is_following_movements:
                    pygame.Rect.__setattr__(self.hitbox, "topleft", (self.hitbox.left + dx, self.hitbox.top + dy))
                    pygame.Rect.__setattr__(self.abs_hitbox, "topleft",
                                            (self.abs_hitbox.left + dx, self.abs_hitbox.top + dy))
                else:
                    self._window.config(offset=(self.window.offset[0] - dx, self.window.offset[1] - dy))
                    self._hitbox = ProtectedHitbox(self.window.get_hitbox())
                    pygame.Rect.__setattr__(self.abs_hitbox, "topleft", (
                        self.parent.abs_rect.left + self.hitbox.left, self.parent.abs_rect.top + self.hitbox.top))
                    pygame.Rect.__setattr__(self.auto_hitbox, "topleft",
                                            (self.hitbox.left - self.rect.left, self.hitbox.top - self.rect.top))
                    if old_hitbox[2:] != self.hitbox.size:
                        pygame.Rect.__setattr__(self.abs_hitbox, "size", self.hitbox.size)
                        pygame.Rect.__setattr__(self.auto_hitbox, "size", self.hitbox.size)
            else:
                pygame.Rect.__setattr__(self.hitbox, "topleft", self.rect.topleft)
                pygame.Rect.__setattr__(self.abs_hitbox, "topleft", self.abs_rect.topleft)

            # We reset the asked_pos after the MOTION in order to allow cycles of pos referecing  TODO
            self._pos_manager._reset_asked_pos()

            if self.is_visible:
                self.send_display_request(rect=self.hitbox.union(old_hitbox))

            self.signal.MOTION.emit(dx, dy)

    def _move_at(self, key, value):

        accepted = (
            "x", "y", "centerx", "centery", "top", "bottom", "left", "right", "topleft", "midtop", "topright",
            "midleft", "center", "midright", "bottomleft", "midbottom", "bottomright")
        assert key in accepted, f"key '{key}' is not a valid rect position (one of {accepted})"

        if self._has_locked.pos:
            return

        old = getattr(self.rect, key)

        if old == value:
            return

        if isinstance(value, (int, float)):

            if key in ("x", "centerx", "left", "right"):
                self._move(value - old, 0)
            else:
                self._move(0, value - old)

        else:
            self._move(value[0] - old[0], value[1] - old[1])

    def _update_pos(self):
        """ Updates the postion from pos_manager's values """

        if self.is_asleep:  # the widget has no parent
            return

        self._move_at(key=self.pos_manager.location, value=self.pos_manager.get_pos_relative_to_owner_parent())

    def collidemouse(self):

        return self.is_visible and self.abs_hitbox.collidepoint(mouse.pos)

    def move(self, dx=0, dy=0):

        if dx == 0 and dy == 0:
            return
        if self._has_locked.pos:
            return
        self._move(dx, dy)

    def send_display_request(self, rect=None):

        if self._parent is not None:  # False when asleep
            if rect is None:
                rect = self.hitbox
            self._parent._warn_change(rect)

    def set_pos(self, **kwarg):
        """ Example : my_widget.set_pos(midtop=(50, 10) """

        assert len(kwarg) == 1
        self._move_at(*kwarg.popitem())

    def set_window(self, window, follow_movements):  # TODO : refed from self
        """window is a rect relative to the parent

        follow_movements default to False"""

        self.window._is_set = window is not None
        if window is None:
            self.window.config(offset=(0, 0), size=self.rect.size)
        else:
            offset_from_parent = window[:2]
            offset_from_self = (
                offset_from_parent[0] - self.rect.left,
                offset_from_parent[1] - self.rect.top
            )
            self.window.config(offset=offset_from_self, size=window[2:])

            self.window._is_following_movements = bool(follow_movements)

            old_pos = self.rect.topleft
            old_size = self.rect.size
            self._hitbox = ProtectedHitbox(self.window.get_hitbox())
            pygame.Rect.__setattr__(self.abs_hitbox, "topleft", (self.parent.abs_rect.left + self.hitbox.left,
                                                                 self.parent.abs_rect.top + self.hitbox.top))
            # NOTE : should auto_hitbox.topleft be (0, 0) or the difference between self.pos and self.window.topleft ?
            pygame.Rect.__setattr__(self.auto_hitbox, "topleft",
                                    (self.hitbox.left - self.rect.left, self.hitbox.top - self.rect.top))
            pygame.Rect.__setattr__(self.abs_hitbox, "size", self.hitbox.size)
            pygame.Rect.__setattr__(self.auto_hitbox, "size", self.hitbox.size)

            if old_pos != self.rect.topleft:
                self.signal.MOTION.emit(self.rect.left - old_pos[0], self.rect.top - old_pos[1])
            if old_size != self.rect.size:
                self.signal.RESIZE.emit(old_size)

        self.send_display_request(rect=self.rect)  # rect covers all possibilities


class HasProtectedSurface(HasProtectedHitbox):

    def __init__(self, parent, surface=None, size=None, **kwargs):
        """NOTE : can be size=(50, 45) or width=50, height=45"""

        HasStyle.__init__(self, parent, options=kwargs)

        self.__asked_size = None
        self._size_hints = [1, 1]
        if size is None:
            style_width = self.style["width"]
            if style_width is None:
                assert surface is not None
                style_width = surface.get_width()
            style_height = self.style["height"]
            if style_height is None:
                assert surface is not None
                style_height = surface.get_height()
            self._asked_size = style_width, style_height
        else:
            self._asked_size = size

        if surface is None:
            surface = pygame.Surface(self._get_asked_size(), pygame.SRCALPHA)

        assert isinstance(surface, pygame.Surface)

        HasProtectedHitbox.__init__(self, parent, surface.get_size(), **kwargs)

        """
        surface is the widget's image

        Duing the widget life, the size of surface can NOT differ from
        the size of the hitbox, it is protected thanks to this classe methods and ProtectedSurface
        
        NEW_SURFACE is emitted right after set_surface()
        """
        self._surface = surface
        self.create_signal("NEW_SURFACE")

        self._dirty = 0
        self._waiting_line = self.parent._children_to_paint

        def update_size_from_askedsize():

            asked_size = self._asked_size
            self.resize(*self._get_asked_size())
            self._asked_size = asked_size

        self.signal.WAKE.connect(update_size_from_askedsize, owner=self)
        self.pos_manager.reference.signal.RESIZE.connect(update_size_from_askedsize, owner=self)

        def check_dirty():
            if self.dirty:
                self._waiting_line.add(self)

        self.signal.WAKE.connect(check_dirty, owner=self)

    surface = property(lambda self: self._surface)

    # HAS_SURFACE
    def _update_size_from_newsurface(self, size):

        with paint_lock:
            pygame.Rect.__setattr__(self.rect, "size", size)
            pygame.Rect.__setattr__(self.abs_rect, "size", size)
            pygame.Rect.__setattr__(self.auto_rect, "size", size)
            if self.window.is_set:
                size = self.window.get_hitbox().size
            pygame.Rect.__setattr__(self.hitbox, "size", size)
            pygame.Rect.__setattr__(self.abs_hitbox, "size", size)
            pygame.Rect.__setattr__(self.auto_hitbox, "size", size)
            self._update_pos()

    def set_surface(self, surface):

        assert isinstance(surface, pygame.Surface), surface

        if self._has_locked.height and self.rect.height != surface.get_height():
            raise PermissionError(
                f"Wrong surface : {surface} (this widget's surface height is locked at {self.rect.h})")

        if self._has_locked.width and self.rect.width != surface.get_width():
            raise PermissionError(
                f"Wrong surface : {surface} (this widget's surface width is locked at {self.rect.w})")

        with paint_lock:
            if self.rect.size != surface.get_size():

                old_hitbox = tuple(self.hitbox)
                old_size = self.rect.size
                self._surface = surface
                self._update_size_from_newsurface(surface.get_size())
                self.signal.RESIZE.emit(old_size)

                if self.is_visible:
                    self.send_display_request(rect=self.hitbox.union(old_hitbox))

            else:

                self._surface = surface
                if self.is_visible:
                    self.send_display_request()

            self.signal.NEW_SURFACE.emit()

    # RESIZABLE
    def _get_asked_size(self):

        size = self._asked_size
        with_percentage = False
        for coord in size:
            if isinstance(coord, str):
                # hard stuff
                assert coord[-1] == '%', size
                with_percentage = True
            else:
                assert isinstance(coord, int), f"Wrong value in size : {coord} (must be a number)"
                assert coord >= 0, f"Wrong value in size : {coord} (must be positive)"

        if with_percentage:
            size = list(size)
            for i, coord in enumerate(size):
                if isinstance(coord, str):
                    size[i] = int(self.parent.rect.size[i] * float(coord[:-1]) / 100)

        return size

    def _set__asked_size(self, new_asked_size):

        self._size_hints: list
        for i, coord in enumerate(new_asked_size):
            if isinstance(coord, str):
                assert coord[-1] == '%', f"Wrong value in size : {coord} (must look like \"75%\")"
                self._size_hints[i] = float(coord[:-1]) / 100
            else:
                assert isinstance(coord, int), f"Wrong value in size : {coord} (must be a number)"
                assert coord >= 0, f"Wrong value in size : {coord} (must be positive)"

        self.__asked_size = new_asked_size

    _asked_size = property(fget=lambda self: self.__asked_size, fset=_set__asked_size)

    # TODO : size_manager ?
    #   size_manager.asked_size
    #   size_manager.size_hints
    #   size_manager.set_asked_size(size)
    #   size_manager.reference

    def _update_surface_from_resize(self, asked_size):
        """ Update the surface from the asked size - Only called by resize()"""

        self.set_surface(pygame.Surface(asked_size, pygame.SRCALPHA))
        self.send_paint_request()

    def resize(self, width, height):
        """Sets up the new widget's surface"""

        if (width, height) == self._asked_size:
            return

        if self.has_locked("width"):
            raise PermissionError("Cannot resize : the width is locked")
        if self.has_locked("height"):
            raise PermissionError("Cannot resize : the height is locked")

        self._asked_size = width, height

        if self.is_asleep:
            return

        asked_size = self._get_asked_size()
        if asked_size == self.rect.size:
            return

        self._update_surface_from_resize(asked_size)

    def resize_height(self, height):

        self.resize(self._asked_size[0], height)

    def resize_width(self, width):

        self.resize(width, self._asked_size[1])

    # PAINTABLE
    dirty = property(lambda self: self._dirty)

    def send_paint_request(self):

        if self._dirty == 0:
            self._dirty = 1

            if self.is_awake:
                self._waiting_line.add(self)

    def set_dirty(self, val):

        assert val in (0, 1, 2)

        self._dirty = val

        if self.is_awake:
            if val:
                self._waiting_line.add(self)
            elif self in self._waiting_line:
                self._waiting_line.remove(self)


class Widget(HasProtectedSurface, metaclass=MetaPaintLocker):

    def __init__(self, parent, layer=None, layer_level=None, name=None, row=None, col=None,
                 visible=True, **kwargs):

        if hasattr(self, "_weakref"):  # Widget.__init__() has already been called
            return

        # name is a string who may help to identify the widget
        # It is defined here, so it's in first place in self.__dict__ (should)
        self._name = name if name else "NoName"

        # INITIALIZATIONS
        HasProtectedSurface.__init__(self, parent, **kwargs)

        self._col = None
        self._row = None
        if (col is not None) or (row is not None):
            """col and row attributes are dedicated to GridLayer"""
            if (self.style["pos"] != (0, 0)) or \
                    (self.style["ref"] is not None) or \
                    (self.style["refloc"] != "topleft"):
                raise PermissionError("When the layer manages the widget's position, "
                                      "all you can define is row, col and loc")
            if col is None:
                col = 0
            if row is None:
                row = 0
            assert isinstance(col, int) and col >= 0
            assert isinstance(row, int) and row >= 0
            self._col = col
            self._row = row

        # LAYOUT
        if layer is None:
            if parent is not self:  # scene prevention
                layer = parent.layers_manager.find_layer_for(self, layer_level)
        elif layer_level is not None:
            raise PermissionError("Cannot define a layer AND a layer_level")
        elif isinstance(layer, str):
            layer = parent.layers_manager.get_layer(layer)
        self._layer = layer

        self._is_visible = visible

    def __repr__(self):
        """
        Called by repr(self)

        :return: str
        """
        # This complicated string creation is avoiding the o.__repr__()
        # in order to avoid representation of widgets
        return f"<{self.__class__.__name__}(name='{self.name}', parent='{self.parent.name}', hitbox={self.hitbox})>"

    def __str__(self):
        """
        Called by str(self)
        """
        return f"{self.__class__.__name__}(name={self.name})"

    # GETTERS ON PROTECTED FIELDS
    col = property(lambda self: self._col)
    layer = property(lambda self: self._layer)
    name = property(lambda self: self._name)
    row = property(lambda self: self._row)

    # ...

    # ...

    # TODO : move these methods to Layer
    def move_behind(self, widget):

        self.layer.move_widget1_behind_widget2(widget1=self, widget2=widget)

    def move_in_front_of(self, widget):

        self.layer.move_widget1_in_front_of_widget2(widget1=self, widget2=widget)

    def swap_layer(self, layer):
        """ NOTE : cannot be called during construction """

        if isinstance(layer, str):
            layer = self.parent.layers_manager.get_layer(layer)
        self.parent.layers_manager.swap_layer(self, layer)



from baopig import *
from baopig.prefabs.colorchooserdialog import ColorChooserDialog


class ThemePresentator(Zone, Selector):
    STYLE = Zone.STYLE.substyle()
    STYLE.modify(
        background_color="theme-color-scene_background",
    )

    def __init__(self, parent, theme):
        Zone.__init__(self, parent, theme=theme, size=("100%", 200))
        Selector.__init__(self, parent)
        Layer(self, padding=10, spacing=10)

        def click():
            ColorChooserDialog(self.application, one_shot=True, theme=self.theme.subtheme()).open()

        Button(self, text="Dialog", command=click)
        Text(self, text="lambda")
        TextEdit(self, text="lambda text in a TextEdit, can you see the ScrollSlider ?", height=40, width=350)
        Slider(self, minval=0, maxval=10, step=1)
        CheckBox(self, "lambda")

        self.pack()


class UT_Theme_Zone(Zone):

    def __init__(self, *args, **kwargs):
        Zone.__init__(self, *args, **kwargs)

        sv = ScrollView(self, size=("100%", "100%"))
        z = Zone(sv, width="100%", padding=(0, 10), spacing=(0, 10))

        ThemePresentator(z, theme="default")
        ThemePresentator(z, theme="dark")
        ThemePresentator(z, theme="pinky")
        ThemePresentator(z, theme="green")

        z.pack(adapt=True)


# For the PresentationScene import
ut_zone_class = UT_Theme_Zone

if __name__ == "__main__":
    from baopig.prefabs.testerscene import TesterScene
    app = Application()
    TesterScene(app, ut_zone_class)
    app.launch()




from baopig import *


class DragableRectangle(Rectangle, DraggableByMouse):
    def __init__(self, parent, **kwargs):
        Rectangle.__init__(self, parent, **kwargs)
        DraggableByMouse.__init__(self, parent, **kwargs)


class DragableZone(Zone, DraggableByMouse):
    def __init__(self, parent, **kwargs):
        Zone.__init__(self, parent, **kwargs)
        DraggableByMouse.__init__(self, parent, **kwargs)


class UT_Origin_Zone(Zone):

    def __init__(self, *args, **kwargs):

        Zone.__init__(self, *args, **kwargs)

        # Magical show
        b = Rectangle(self, pos=(7, 7), ref=self, color=(0, 0, 0, 0), touchable=False,
                      border_color=(128, 10, 10), size=(100, 100), border_width=3)

        # Prisonner at center of the magical show
        z = DragableZone(self, sticky="midtop", size=(100, 100), background_color=(40, 34, 34), name="bottom")
        z.move_behind(b)
        DragableRectangle(z, color=(0, 128, 128), size=(30, 30), center=z.auto_rect.center, ref=b)

        # Prisonner inside magical show, at center of its zone
        z = DragableZone(self, sticky="topright", size=(100, 100), background_color=(40, 34, 34), name="bottomright")
        z.move_behind(b)
        z_window = Zone(z, size=("100%", "100%"), ref=b)
        DragableRectangle(z_window, color=(0, 128, 128), size=(30, 30), center=z.auto_rect.center, ref=z)

        # CLOCK
        z2 = DragableZone(self, midtop=("-50%", 186), refloc="topright",
                          size=(350, 350), background_color=(140, 140, 140), name="z3")
        z3 = DragableZone(z2, sticky="center",
                          size=(250, 250), background_color=(150, 150, 150), name="z2")
        ref = DragableRectangle(z3, center=("50%", "50%"), ref=self,
                                color=(128, 128, 0), size=(30, 30), name="ref")
        import math
        radius = 100
        step = int(math.degrees((2 * math.pi) / 8))
        for i in range(0, int(math.degrees(2 * math.pi)), step):
            Rectangle(z2, color=(150, 120, 0), size=ref.rect.size,
                      pos=(math.cos(math.radians(i)) * radius,
                           math.sin(math.radians(i)) * radius),
                      ref=ref, name="rect({})".format(i))
        for i in range(0, int(math.degrees(2 * math.pi)), step):
            i += step / 2
            Rectangle(z3, color=ref.color, size=ref.rect.size,
                      pos=(math.cos(math.radians(i)) * radius,
                           math.sin(math.radians(i)) * radius),
                      ref=ref, name="rect({})".format(i))

        # Family
        r1 = DragableRectangle(self, pos=(0, "50%"), color=(100, 50, 25), size=(30, 30), name="r1")
        r2 = DragableRectangle(self, pos=(40, 0), ref=r1, color=(50, 100, 25), size=(30, 30), name="r2")
        r3 = DragableRectangle(self, pos=(0, 40), ref=r2, color=(100, 50, 25), size=(30, 30), name="r3")
        DragableRectangle(self, pos=(-40, 0), ref=r3, color=(50, 100, 25), size=(30, 30), name="r4")
        DragableRectangle(self, pos=("50%", "50%"), ref=r1, color=(75, 75, 25), size=(40, 40))

        # CROSS
        Rectangle(self, color=(0, 0, 0), size=(6, 20), sticky="center")
        Rectangle(self, color=(0, 0, 0), size=(20, 6), sticky="center")

    def load_sections(self):
        self.parent.add_section(
            title="Selector",
            tests=[
                "the cross is always at the texting zone center, even after a resize",
                "the topright and midtop corners can be dragged",
                "the topright and midtop corners follow the testing zone resizing even after being moved",
                "if the topright corner is set inside the topleft border, a prisonner appears at corner's center",
                "if the midtop corner is set inside the topleft border, a prisonner appears at border'center",
                "when the midtop corner moves, the prisonner is visually static, even with low fps",
                "a prisonner can be dragged",
                "the clock (light gray surface at the center) can be dragged",
                "after the scene width changed, the clock abcissa is still at the center of the application",
                "if the clock has moved, after a scene resizing, it keeps the same distance from the scene's right",
                "the belt (rects around the clock center) cannot be dragged",
                "the yellow center (yellow rect at the clock center) can be dragged",
                "the belt follow the yellow center everywhere he goes",
                "dragging the yellow center don't cause lag",
            ]
        )


# For the PresentationScene import
ut_zone_class = UT_Origin_Zone

if __name__ == "__main__":
    from baopig.prefabs.testerscene import TesterScene
    app = Application()
    TesterScene(app, ut_zone_class)
    app.launch()





from baopig import *


class SelectableRectangle(Rectangle, SelectableWidget):

    def __init__(self, parent, pos):
        Rectangle.__init__(self, parent, color="blue", size=(30, 30), pos=pos)
        SelectableWidget.__init__(self, parent)
        self._hightlighter_ref = Highlighter(self.parent, self, border_width=2, layer_level=2).get_weakref()
        self.hightlighter.sleep()
        self.timer = Timer(.4, self.hightlighter.sleep)

    hightlighter = property(lambda self: self._hightlighter_ref())

    def handle_select(self):
        self.set_color("blue4")
        self.highlight("green")

    def handle_unselect(self):
        self.set_color("blue")
        self.highlight("red")

    def highlight(self, color):
        self.timer.cancel()
        self.hightlighter.wake()
        self.hightlighter.set_border_color(color)
        self.timer.start()


class SelectorZone(Zone, Selector):

    def __init__(self, parent, can_select=True, **kwargs):
        Zone.__init__(self, parent, **kwargs)
        Selector.__init__(self, parent, can_select)


class UT_Selections_Zone(SelectorZone):

    def __init__(self, *args, **kwargs):
        SelectorZone.__init__(self, *args, **kwargs)

        z = SelectorZone(self, size=(int(self.rect.w / 3), self.rect.h - 20), background_color="gray",
                         midtop=("50%", 10))
        z.set_style_for(SelectionRect, border_color="red", color=(255, 0, 0, 40))
        SelectableRectangle(z, (10, 10))
        SelectableRectangle(z, (50, 10))
        SelectableRectangle(z, (90, 10))
        Text(z, "I am selectable", pos=(10, 50))
        TextEdit(z, max_width=z.rect.w - 20, pos=(10, 75))  # TODO : Scrollable

        z2 = SelectorZone(z, size=(z.rect.w - 20, int((z.rect.h - 40) / 3)), background_color=(128, 128, 128),
                          midleft=(10, "50%"), can_select=False)
        SelectableRectangle(z2, (10, 10))
        SelectableRectangle(z2, (50, 10))
        SelectableRectangle(z2, (90, 10))
        Text(z2, "I am not selectable", pos=(10, 50))
        TextEdit(z2, max_width=z2.rect.w - 20, pos=(10, 75))

        z3 = SelectorZone(z, size=(z.rect.w - 20, int((z.rect.h - 40) / 3)), background_color=(128, 128, 128, 200),
                          pos=(0, -10), sticky="midbottom")
        z3.set_selectionrect_visibility(False)
        SelectableRectangle(z3, (10, 10))
        SelectableRectangle(z3, (50, 10))
        SelectableRectangle(z3, (90, 10))
        Text(z3, "Selection rectangle ?", pos=(10, 50))
        TextEdit(z3, max_width=z3.rect.w - 20, pos=(10, 75))

        SelectableRectangle(self, (10, 10))
        SelectableRectangle(self, (50, 10))
        SelectableRectangle(self, (90, 10))
        Text(self, "I am selectable", pos=(10, 50))
        TextEdit(self, max_width=z.rect.w - 20, pos=(10, 75))

        SelectableRectangle(self, (z.rect.right + 10, 10))
        SelectableRectangle(self, (z.rect.right + 50, 10))
        SelectableRectangle(self, (z.rect.right + 90, 10))
        Text(self, "I am selectable", pos=(z.rect.right + 10, 50))
        TextEdit(self, max_width=z.rect.w - 20, pos=(z.rect.right + 10, 75))

    def load_sections(self):
        self.parent.add_section(
            title="Selector",
            tests=[
                "A Selector can be focused",
                "When receive Ctrl + A, calls the 'select_all' method",
                "The default implementation of 'select_all' selects all selectable children",
                "When receive link_motion, it creates a SelectionRect",
                "The default implementation of 'copy' collects data from selected Selectables, join into a string "
                "and send it to the clipboard",
                "When receive Ctrl + V, calls the 'paste' method (no default implementation)",
                "The call of 'disable_selecting' will deactivate the selection ability",
            ]
        )

        self.parent.add_section(
            title="SelectableWidget",
            tests=[
                "A SelectableWidget's selector is the youngest Selector in SelectableWidget's family tree",
                "For each selection rect movement, if it collide with a selectable, it calls the 'select' method",
                "At the time selection rect don't collide with a selectable anymore, 'unselect' is called",
            ]
        )

        self.parent.add_section(
            title="The selection rectangle",
            tests=[
                "A drag on a focused Selector create a selection rect",
                "A released drag will hide the selection rect",
                "The selection rect can be configured throught passing a subclass of SelectionRect "
                "in the Selector constructor",
                "The visibility of the selection rect can be edited -> set_selectionrect_visibility",
                "When setting an end position for the selection rect, a temporary visibility can be given in argument",
                "The selection rect is always fitting inside its parent",
                "The middle zone's selection rect is red",
            ]
        )


# For the PresentationScene import
ut_zone_class = UT_Selections_Zone

if __name__ == "__main__":
    from baopig.prefabs.testerscene import TesterScene
    app = Application()
    TesterScene(app, ut_zone_class)
    app.launch()





import time
import pygame
from baopig.pybao.objectutilities import Object, History
from baopig.communicative import Communicative
from baopig.documentation import Container, Focusable, HoverableByMouse, LinkableByMouse, ScrollableByMouse
from .logging import LOGGER


class _Mouse(Communicative):

    _signals = "MOUSEMOTION", "MOUSEBUTTONDOWN", "MOUSEBUTTONUP", "MOUSEWHEEL"

    def __init__(self):

        Communicative.__init__(self)

        # 4 signals :
        #     - MOUSEMOTION
        #     - MOUSEBUTTONDOWN
        #     - MOUSEBUTTONUP
        #     - MOUSEWHEEL
        for signal in self._signals:
            self.create_signal(signal)

        self._pos = (-1, -1)  # No collision at application launch
        self._rel = (0, 0)  # Le dernier deplacement de la souris

        # L'etat des bouttons de la souris (il y en a 5)
        # self.button[3] -> is the button 3 pressed ?
        # There is no button 0, so self.button[0] = None
        # For pygame, the 4 and 5 buttons are wheel up and down,
        # so these buttons are implemented as mouse.SCROLL

        self._pressed_buttons = {}

        # [None, 0, 0, 0]  # WARNING : A mouse can have additionnals buttons

        # Using an empty clic avoid testing if the mouse has clicks in memory or not
        empty_clic = Object(
            time=-1,  # heure du clic - en secondes
            button=None,  # numero du bouton
            pos=None)  # position du clic
        historic_size = 3  # Pour le triple-clic
        self.clic_history = History(maxlen=historic_size, seq=[empty_clic] * historic_size)

        """
        When the mouse is hovering a Text inside a Button inside a Zone inside a Scene,
        then the Text is hovered
        """
        self._hovered_widget = None

        """
        When the mouse click on a Text inside a Button inside a Zone inside a Scene,
        then the Text is linked
        """
        self._linked_widget = None

        """
        Permet de savoir si l'utilisateur vient de faire un double-click

        Pour faire un double-clic, il faut :
            - qu'il y ait eu 2 clics en moins de 5 dixiemes de secondes
            - que les 2 clics aient ete fait par le bouton gauche
            - que la souris n'ait pas bougee entre les clics

        has_double_clicked garde la valeur True jusqu'au clic suivant
        """
        self.has_double_clicked = False

        """
        Permet de savoir si l'utilisateur vient de faire un triple-clic

        Pour faire un triple-clic, il faut :
            - qu'il y ait eu 3 clics en moins de 10 dixiemes de secondes
            - que les 3 clics aient ete fait par le bouton gauche
            - que la souris n'ait pas bougee entre les clics

        has_triple_clicked garde la valeur True jusqu'au clic suivant
        """
        self.has_triple_clicked = False

        self._application = None
        self._display = None
        self._is_hovering_display = True

    def __repr__(self):
        return "<Mouse(" + str(self.__dict__) + ")>"

    def __str__(self):
        return f"<Mouse(pos={self.pos}, pressed_buttons={self._pressed_buttons})>"

    pos = property(lambda self: self._pos)
    x = property(lambda self: self._pos[0])
    y = property(lambda self: self._pos[1])

    is_hovering_display = property(lambda self: self._is_hovering_display)
    linked_widget = property(lambda self: self._linked_widget)
    hovered_widget = property(lambda self: self._hovered_widget)

    def _link(self, widget):

        assert self.is_hovering_display
        assert self.linked_widget is None

        if widget is not None:
            assert not widget.is_linked

            self._linked_widget = widget
            widget.is_linked = True

            widget.signal.HIDE.connect(self._unlink, owner=None)
            widget.signal.SLEEP.connect(self._unlink, owner=None)
            widget.signal.KILL.connect(self._unlink, owner=None)

            self.linked_widget.handle_link()

    def _get_touched_widget(self):
        """ Return the youngest touchable widget that is touched """

        def get_touched_widget(cont):

            for layer in reversed(tuple(cont.layers_manager.touchable_layers)):
                assert layer.touchable
                for child in reversed(layer):
                    if child.is_touchable_by_mouse and child.collidemouse():
                        if isinstance(child, Container):
                            touched = get_touched_widget(child)
                            if touched is not None:
                                return touched
                        return child
            if cont.is_touchable_by_mouse:
                return cont

        return get_touched_widget(self._application.focused_scene)

    def _hover_display(self):

        if self.is_hovering_display:
            return
        self._is_hovering_display = True
        self.update_hovered_widget()

    def _unhover_display(self):

        if not self.is_hovering_display:
            return
        self._hover(None)
        self._is_hovering_display = False

    def _hover(self, widget):

        if widget is self._hovered_widget:
            return

        # UNHOVER
        old_hovered = self._hovered_widget
        if self._hovered_widget is not None:
            assert old_hovered.is_hovered
            old_hovered._is_hovered = False

            old_hovered.signal.HIDE.disconnect(self.update_hovered_widget)
            old_hovered.signal.SLEEP.disconnect(self.update_hovered_widget)
            old_hovered.signal.KILL.disconnect(self.update_hovered_widget)

        self._hovered_widget = widget

        # HOVER
        if widget is not None:
            assert widget.is_visible, repr(widget)

            assert not widget.is_hovered
            widget._is_hovered = True

            widget.signal.HIDE.connect(self.update_hovered_widget, owner=None)
            widget.signal.SLEEP.connect(self.update_hovered_widget, owner=None)
            widget.signal.KILL.connect(self.update_hovered_widget, owner=None)

        # SIGNALS
        if old_hovered is not None:
            old_hovered.signal.UNHOVER.emit()

        if widget is not None:
            widget.signal.HOVER.emit()

    def _unlink(self):
        """
        This method unlinks a LinkableByMouse widget from the mouse

        It can exceptionnaly be called when a clicked widget disappears
        Then the widget calls itself this function, trougth LinkableByMouse.unlink()
        """

        try:
            assert self.linked_widget.is_linked or self.linked_widget.is_dead
        except AssertionError as e:
            raise e

        widget = self.linked_widget
        self._linked_widget = None

        if widget.is_alive:
            widget.is_linked = False

            widget.signal.HIDE.disconnect(self._unlink)
            widget.signal.SLEEP.disconnect(self._unlink)
            widget.signal.KILL.disconnect(self._unlink)

            widget.handle_unlink()
        # While the mouse left button was press, we didn't update hovered_widget
        self.update_hovered_widget()

    def get_pos_relative_to(self, widget):

        return widget.abs_rect.referencing(self.pos)

    def is_pressed(self, button_id):
        """Return True if the button with identifier 'button_id' (an integer) is pressed"""

        try:
            return bool(self._pressed_buttons[button_id])
        except KeyError:
            # Here, the button has never been pressed
            return 0

    def receive(self, event):

        # Unknown & skipable events
        if event.type not in (pygame.MOUSEBUTTONUP, pygame.MOUSEBUTTONDOWN, pygame.MOUSEMOTION):
            LOGGER.warning(f"Unknown event : {event}")
            return
        if event.type in (pygame.MOUSEBUTTONUP, pygame.MOUSEBUTTONDOWN):

            if event.button not in (1, 2, 3, 4, 5):
                LOGGER.warning(f"Unknown button id : {event.button} (event : {event})")
                return

            # if self._pressed_buttons[event.button] is False:
            # if self.pressed_button is not None and self.pressed_button != event.button:
            #     # Another button is already pressed, we skip
            #     LOGGER.info("Another button is already pressed, we skip")
            #     return

            if event.type == pygame.MOUSEBUTTONUP:
                if event.button in (4, 5):
                    # It's a wheel end of scrolling, wich is useless
                    return

                # if not self.is_pressed(event.button):
                #     # This button's DOWN event have been skiped, so we skip it's UP event
                #     return
        elif event.type == pygame.MOUSEMOTION:

            if event.rel == (0, 0):
                # Don't care
                return

        # Real work

        if event.type == pygame.MOUSEBUTTONDOWN:

            # ACTUALIZING MOUSE STATE

            if event.button in (4, 5):

                def first_scrollable_in_family_tree(widget):
                    if widget.parent == widget:
                        return None
                    if isinstance(widget, ScrollableByMouse) and widget.is_touchable_by_mouse:
                        return widget
                    return first_scrollable_in_family_tree(widget.parent)

                pointed = self._get_touched_widget()
                scrolled = first_scrollable_in_family_tree(pointed)
                if scrolled:
                    scrolled.handle_mouse_scroll(event)

            else:  # Ignore wheel
                self.clic_history.append(Object(time=time.time(), button=event.button, pos=event.pos))
                self._pressed_buttons[event.button] = True

                self.has_double_clicked = \
                    self.clic_history[-2].time > self.clic_history[-1].time - .5 and \
                    self.clic_history[-2].button == self.clic_history[-1].button == 1 and \
                    self.clic_history[-2].pos == self.clic_history[-1].pos

                self.has_triple_clicked = \
                    self.has_double_clicked and \
                    self.clic_history[-3].time > self.clic_history[-1].time - 1 and \
                    self.clic_history[-3].button == self.clic_history[-1].button == 1 and \
                    self.clic_history[-3].pos == self.clic_history[-1].pos

            if event.button == 1:
                # UPDATING CLICKS, FOCUSES, HOVERS...

                def first_linkable_in_family_tree(widget):
                    # The recursivity always ends because a Scene is a LinkableByMouse
                    if isinstance(widget, LinkableByMouse) and widget.is_touchable_by_mouse:
                        return widget
                    return first_linkable_in_family_tree(widget.parent)

                def first_focusable_in_family_tree(widget):
                    # The recursivity always ends because a Scene is a Focusable
                    if isinstance(widget, Focusable) and widget.is_touchable_by_mouse:
                        return widget
                    return first_focusable_in_family_tree(widget.parent)

                pointed = self._get_touched_widget()
                linked = first_linkable_in_family_tree(pointed)
                focused = first_focusable_in_family_tree(linked)
                # Le focus passe avant le link
                self._application.focused_scene.focus(focused)
                self._link(linked)

        elif event.type == pygame.MOUSEBUTTONUP:

            assert self.is_pressed(event.button)

            # ACTUALIZING MOUSE STATE
            self._pressed_buttons[event.button] = False

            # Linkable stuff
            if self.linked_widget:
                self._unlink()

        elif event.type == pygame.MOUSEMOTION:

            # ACTUALIZING MOUSE STATE
            self._pos = event.pos

            # Linkable and Hoverable stuff
            if self.is_pressed(button_id=1) and self.linked_widget:
                self.linked_widget.handle_link_motion(event.rel)
            else:
                self.update_hovered_widget()

        # EVENT TRANSMISSION
        for signal_id in self._signals:
            if event.type == getattr(pygame, signal_id):
                getattr(self.signal, signal_id).emit(event)
                break

    def update_hovered_widget(self):

        if self.linked_widget is not None:
            return

        def first_hoverable_in_family_tree(widget):
            if widget.scene is widget:
                if isinstance(widget, HoverableByMouse):
                    return widget
                return None
            if isinstance(widget, HoverableByMouse):
                return widget
            else:
                return first_hoverable_in_family_tree(widget.parent)

        pointed = self._get_touched_widget()
        if pointed is None:
            self._hover(None)
        else:
            self._hover(first_hoverable_in_family_tree(pointed))


mouse = _Mouse()
del _Mouse



from typing import Iterable

import pygame


class Communicative:
    """
    Class for objects who need to emit signals

    :Example:
    ---------
        class Transmitter(bp.Communicative):
            def __init__(self, message):
                bp.Communicative.__init__(self)
                self.message = message
                self.create_signal("EMISSION")
            def transmit(self):
                self.signal.EMISSION.emit(self.message)

        class Listener(bp.Communicative):
            def receive(self, message):
                print(f"Message received: {message}")

        my_transmitter = Transmitter("Hello world !")
        my_listener = Listener()
        my_transmitter.signal.EMISSION.connect(my_listener.receive, owner=my_listener)
        my_transmitter.transmit()  # prints : 'Message received: Hello world !'
        my_listener.disconnect()
        my_transmitter.transmit()  # prints nothing

    :Attributes:
    ------------
        signal: object -> an object containing all the signals

    :Methods:
    ---------
        create_signal(signal_id) -> creates a signal accessible via 'signal_id'
        disconnect()             -> kills all the connections owned by this object
    """

    def create_signal(self, signal_id: str):
        """
        Creates a signal accessible via 'signal_id'
        'signal_id' must be an uppercase string

        After obj.create_signal("EMISSION"), the created signal is accessed this way:
            obj.signal.EMISSION
        """

    def disconnect(self):
        """ Kills all the connections owned by this object """


class ProtectedRect:
    pass


class Widget(Communicative):
    """
    Abstract class for the graphical elements of the screen
    A widget can be visible, hidden, awake, sleeping, alive, or dead
    A widget has only one parent (baopig bio-logic, hmmm...)
    A widget has to initialize its surface at its creation

    WARNING : A widget's surface is not connected to its widget (that is way too hard)
              Any change on the surface has to come with :
                  - a signal emission : widget.signal.NEW_SURFACE.emit()
                  - a display request : widget.send_display_request()
              This is not required when you use the set_surface() method

    :Constructor:
    -------------
        parent: Container       -> the widget's parent
        surface: pygame.Surface -> the widget's image
        visible: bool           -> if False, the widget's hide() method is called
        sticky: Location        -> if set, overrides 2 style attributes : pos_loc & refloc
        **style: keyword args   -> the widget's style attributes

    :Signals:
    ---------
        HIDE: emitted when the widget's visibility state goes from visible to hidden
        KILL: emitted when the widget gets killed
        SHOW: emitted when the widget's visibility state goes from hidden to visible
        SLEEP: emitted when the widget's sleeping state goes from awake to asleep
        WAKE: emitted when the widget's sleeping state goes from asleep to awake

    :Attributes:
    ------------

        :Style attributes:
        -----------------------
            pos: Iterable[int]         -> the position
            loc: Location              -> the position's location on the widget's rect
            ref: Container | None      -> the position's position reference, if None, set to parent
            refloc: Location           -> the position's position reference location, from the reference's rect
            referenced_by_hitbox: bool -> if True, position is referenced by the ref's hitbox  # TODO : tests

        parent: Container -> the manager

        surface: pygame.Surface -> the widget's image
        rect: pygame.Rect       -> the widget's size
        hitbox: pygame.Rect     -> the widget's size on the screen
        pos_manager: object     -> the widget's position manager. See documentation at TODO

        is_alive: bool   -> True if the widget has not been killed
        is_asleep: bool  -> True if the widget is asleep
        is_awake: bool   -> True if the widget is not asleep
        is_dead: bool    -> True if the widget has been killed
        is_hidden: bool  -> True if the widget is not visible
        is_visible: bool -> True if the widget is visible

    :Methods:
    ---------
        set_pos(pos=None, pos_loc=None, refloc=None) -> moves the widget  # TODO
            # pos = pos + ref.abs_rect.refloc - parent.abs_rect.topleft
            # rect.pos_loc = pos
            # signal.MOTION.emit()

        set_size(width, height) -> resizes the widget  # TODO
            # rect.size = width, height
            # signal.RESIZE.emit()

        set_surface() TODO

        hide()  -> the widget is no longer visible
        kill()  -> the widget is permanently deleted
        show()  -> the widget is visible again
        sleep() -> the widget is detached from its parent
        wake()  -> the widget is reattached to its parent

        send_display_request(rect=None) -> sends a request who will update the display
    """

    application: ...
    is_alive: bool
    is_asleep: bool
    is_awake: bool
    is_dead: bool
    is_hidden: bool
    is_touchable_by_mouse: bool
    is_visible: bool
    parent: ...
    pos_manager: ...
    scene: ...
    surface: pygame.Surface

    # HITBOX
    abs_hitbox: ProtectedRect
    abs_rect: ProtectedRect
    auto_hitbox: ProtectedRect
    auto_rect: ProtectedRect
    hitbox: ProtectedRect
    rect: ProtectedRect
    window: ProtectedRect

    def hide(self):
        """
        Stop displaying the widget

        Behaviour:
        ----------
            Sets visibility to False
            Sends a display request
            Emits the HIDE signal
        """

    def kill(self):
        """
        Kills the widget
        A killed widget should not be used ever again

        Behaviour:
        ----------
            Detaches the widget from its parent
            Emits the KILL signal
            Kills all the connections owned by the widget
            Kills the widget's weakref
        """

    def send_display_request(self, rect: Iterable[int] | None = None):
        """
        Sends a request who will update the display
        The request is executed by a thread dedicated to the screen's display
        rect represents a pygame.Rect object
        rect must be referenced by the widget's parent
        If rect is not set, the widget's hitbox will be used
        Only the screen's portion corresponding to rect will be updated
        """

    def show(self):
        """
        Starts to display the widget again

        Behaviour:
        ----------
            Sets visibility to True
            Sends a display request
            Emits the SHOW signal
        """

    def sleep(self):
        """
        Detaches the widget from its parent
        A sleeping widget cannot move

        Behaviour:
        ----------
            Detaches the widget from its parent
            Emits the SLEEP signal
        """

    def wake(self):
        """
        Reattaches the widget to its parent

        Behaviour:
        ----------
            If the parent got killed:
                Kills the widget
            Attaches the widget to its parent
            Updates the widget's size & position
            Emits the WAKE signal
        """

    # TODO : merge
    """
    Class for widgets who may need to update their surface.

    Attribute dirty means the widget's surface has to be updated.
    The widget's surface is updated through the paint() method.
    The paint() method is executed by a thread dedicated to the screen's display.
    If an asleep widget is dirty, the paint() method is called when the widget awakes.
    WARNING : It is deprecated to call paint() yourself, use send_paint_request() instead.

    :Attributes:
    ------------
        dirty: int
            -> if 0, paint() is not requested
            -> if 1, paint() is called at next frame rendering, then dirty will be set to 0 again
            -> if 2, paint() is called at each frame rendering

    :Methods:
    ---------
        paint()              -> abstract - updates the widget's surface
        send_paint_request() -> sends a request who will execute once the widget's paint() method
        set_dirty(val)       -> sets the widget's dirty attribute
    """
    def paint(self):
        """
        Abstract - called at a frame rendering, if the widget is dirty.

        WARNING : NEVER CALL paint() YOURSELF !! Use send_paint_request() instead.

        When called by a paint_request, paint() is followed by these 2 events:
            - widget's signal NEW_SURFACE is emitted
            - a display_request is sent

        :Examples:

            def paint(self):
                surf = get_the_new_surface()
                self.set_surface(surf)

            def paint(self):
                self.surface.blit(another_surface, (0, 0))
        """

    def send_paint_request(self):
        """
        Sends a request who will execute once the widget's paint() method
        The request is executed by a thread dedicated to the screen's display
        Acts almost like set_dirty(1)
        """

    def set_dirty(self, val: int):
        """
        Sets the widget's dirty attribute

        Paintable.dirty:
            if 0, paint() is not requested
            if 1, paint() is called at next frame rendering, then dirty will be set to 0 again
            if 2, paint() is called at each frame rendering
        """


class HoverableByMouse(Widget):
    """
    Class for widgets who need to handle when they are hovered or unhovered by the mouse.

    A HoverableByMouse can have an Indicator, i.e. a Text that appears when the HoverableByMouse is hovered.
    For more details, see TODO : link to Indicator documentation

    TODO : test : a disabled HoverableByMouse cannot be hovered
    TODO : test : an enabled HoverableByMouse can grab the hover
    TODO : test : a disabled & hovered HoverableByMouse drops the hover

    :Signals:
    ---------
        HOVER: emitted when the widget gets hovered by the mouse
        UNHOVER: emitted when the widget gets unhovered from the mouse

    :Attributes:
    ------------
        is_hovered: bool     -> True if the mouse is hovering the widget
        indicator: Indicator -> the widget's indicator, not always set
        # TODO : why is there a ref ? cannot have multiple indicators ?

    :Methods:
    ---------
        handle_hover()   -> abstract - called when the widget gets hovered by the mouse
        handle_unhover() -> abstract - called when the widget gets unhovered from the mouse
    """

    def handle_hover(self):
        """ Abstract - called when the widget gets hovered by the mouse """

    def handle_unhover(self):
        """ Abstract - called when the widget gets unhovered from the mouse """


class Paintable(Widget):  # TODO : remove
    pass


class Runable(Widget):
    """
    Class for widgets who need to execute their run() method as much as possible.

    By default, is_running is set to False.

    :Attributes:
    ------------
        is_running: bool -> True if the widget is running

    :Methods:
    ---------
        run() -> abstract - called as much as possible, while the object is running

        set_running(val) -> starts or stops to run the widget
    """

    is_running: bool

    def run(self):
        """ Abstract - called as much as possible, while the object is running """

    def set_running(self, val):
        """ Starts or stops to run the widget """


# ...


"""

                                      Widget   -> set_touchable_by_mouse()
                                 
                                         |
                                 
                                 HoverableByMouse
                                 
                                         |
                                 
                                 LinkableByMouse       
                                 
                                          \
                                 
                                           Focusable  ( KeyEventsListener )
                                 
                                                \     
                                                 \    
                    Clickable                     \     
                                                   \  
                                                    \ 
                                         --------------------------------------
                                         |                                    |
                                     
                                       Button                             Selector
                                       
                                                                              |
                                                                              
                                                                          TextEdit
                                                                          LineEdit
                                                                          Entry
                                                                          NumEntry
                                                                          ColorEntry


"""


class Validable:
    pass


class ScrollableByMouse(Widget):

    def handle_mouse_scroll(self, scroll_event):
        """TODO"""


class LinkableByMouse(HoverableByMouse):
    """
    Class for widgets who need to capture mouse clicks

    A LinkableByMouse is linked when a mouse LEFT BUTTON DOWN collide with it,
    and unlinked when the LEFT BUTTON UP occurs

    It has no 'link' or 'link_motion' method since it is the mouse who manages links.
    However, it can unlink itself

    WARNING : If a LinkableByMouse parent contains a LinkableByMouse child, and the LEFT BUTTON
              DOWN has occured on the child, then only the child will be linked

    NOTE : when a widget is linked, you can access it via mouse.linked_widget
    """


class Focusable(LinkableByMouse):
    """
    Class for widget who listen to keyboard input, when it has the focus. Only one
    widget can be focused in the same time. When a new clic occurs, and it doesn't collide
    this widget, it is defocused.

    It has no 'focus' method since it is the application who decide who is focused or not.
    However, it can defocus itself.

    When the mouse click on a Text inside a Button inside a focusable Zone inside a Scene,
    then only the youngest Focusable is focused
    Here, it is the Button -> scene.focused_widget = Button
    """

    def handle_keydown(self, key: int):
        """ Called when a key is pressed """

    def handle_keyup(self, key: int):
        """ Called when a key is released """


class Container(Widget):
    """
    Widgets parent

    Attributes:
    -----------
        children: list -> the list of all the children
    """

    layers_manager: ...


class Selector(Container, Focusable):
    """
    Class for containers who need to handle when they are linked
    and then, while the mouse is still pressed, to handle the mouse drag in
    order to simulate a rect from the link origin to the link position and
    select every SelectableWidget object who collide with this rect
    """


class Layout(Container):
    """
    Widgets's size and position manager

    Attributes:
    -----------
        padding: Margin
        spacing: Margin

        content_rect: pygame.Rect -> the space for children

    """


class Zone(Container):
    """
    Convenient Widgets manager

    Attributes:
    -----------
        background: Layer
        content: Layer
        foreground: Layer
    """


class Scene(Zone):
    """
    An application frame

    Methods:
        open
    """


class ApplicationExit(Exception):
    pass



from .style import Theme
from .utilities import *
from .widget import Widget, Communicative
from .widget_supers import DraggableByMouse, Focusable, HoverableByMouse, LinkableByMouse, \
    RepetivelyAnimated, Runable, Validable
from .imagewidget import Image
from .shapes import Rectangle, Highlighter, Polygon, Line, Circle
from .layer import Layer
from .gridlayer import GridLayer
from .layersmanager import LayersManager
from .container import Container
from .zone import Zone, SubZone
from .selections import SelectableWidget, Selector, SelectionRect
from .scene import Scene
from .application import Application



import time
import pygame
from baopig.documentation import ScrollableByMouse as ScrollableByMouseDoc
from baopig.io import keyboard
from baopig.lib import Rectangle
from baopig.lib import Zone, LayersManager
from .slider import Slider, SliderBar, SliderBloc


class ScrollBloc(SliderBloc):
    STYLE = SliderBloc.STYLE.substyle()
    STYLE.modify(
        wideness=10,
        border_width=0,
        color=(100, 100, 100, 63)
    )

    def __init__(self, slider):

        SliderBloc.__init__(self, slider)

        slider.signal.NEW_VAL.connect(self.update_length, owner=self)

    def update_length(self):

        try:
            if self.slider.axis == "x":
                self.resize_width(int(self.slider.rect.width ** 2 / self.slider.parent.main_widget.rect.width))
            else:
                self.resize_height(int(self.slider.rect.height ** 2 / self.slider.parent.main_widget.rect.height))
        except ZeroDivisionError:  # main_widget's length is null
            pass


class ScrollBar(SliderBar):
    STYLE = SliderBar.STYLE.substyle()
    STYLE.modify(
        wideness=10,
        border_width=0,
        color=(0, 0, 0, 0)
    )


class ScrollSlider(Slider):
    STYLE = Slider.STYLE.substyle()
    STYLE.modify(
        bloc_class=ScrollBloc,
        bar_class=ScrollBar,
        has_indicator=False,
        wideness=10,
    )

    def __init__(self, scroller, axis):

        assert isinstance(scroller, ScrollView)

        self._parent = scroller
        self._axis = axis

        if axis == "x":
            pos = (0, 0)
            sticky = "midbottom"
            length = scroller.rect.width
        else:
            pos = (0, 0)
            sticky = "midright"
            length = scroller.rect.height

        Slider.__init__(
            self, scroller, step=1e-9, axis=axis, pos=pos, sticky=sticky, layer_level=LayersManager.FOREGROUND,
            minval=0, maxval=scroller.max[axis], length=length,
        )

        self.signal.NEW_VAL.connect(self.handle_new_val, owner=None)

        self._hover_sail_ref = Rectangle(self, size=("100%", "100%"), touchable=False, color=(255, 255, 255, 63),
                                         visible=False).get_weakref()

        if not self.range:
            self.hide()

    hover_sail = property(lambda self: self._hover_sail_ref())

    def handle_hover(self):

        self.hover_sail.show()

    def handle_unhover(self):

        self.hover_sail.hide()

    def handle_new_val(self):

        self.parent._set_scrollval(self.axis, self.val)

    def set_val(self, val):

        self._update_val(val=val)

    def update_from_mainwidget_resize(self):

        maxval = self.parent.max[self.axis]

        if maxval == self.maxval:
            return

        self._range = self._maxval = maxval

        if self.axis == "x":
            val = - self.parent.main_widget.rect.left + self.parent.padding.left
        else:
            val = - self.parent.main_widget.rect.top + self.parent.padding.top

        if val == self.val:
            self.bloc.update()
            self.handle_new_val()
        else:
            self.set_val(val)
        self.bloc.update_length()

        if maxval:
            self.show()
        else:
            self.hide()


class ScrollView(ScrollableByMouseDoc, Zone):
    STYLE = Zone.STYLE.substyle()
    STYLE.create(
        scrollslider_class=ScrollSlider
    )

    def __init__(self, parent, **kwargs):

        Zone.__init__(self, parent, **kwargs)

        self.max = {"x": 0, "y": 0}
        self._last_scroll_time = 0

        scrollslider_class = self.style["scrollslider_class"]
        self._x_scroller_ref = scrollslider_class(self, axis="x").get_weakref()
        self._y_scroller_ref = scrollslider_class(self, axis="y").get_weakref()

        self._main_widget = None

    main_widget = property(lambda self: self._main_widget())
    x_scroller = property(lambda self: self._x_scroller_ref())
    y_scroller = property(lambda self: self._y_scroller_ref())

    def _add_child(self, child):

        super()._add_child(child)
        if child.layer.level == LayersManager.MAINGROUND:
            assert self._main_widget is None, "A ScrollView cannot contain more than one widget"
            assert child.pos_manager.reference is self
            assert child.pos_manager.location == "topleft"
            assert child.pos_manager.reference_location == "topleft"
            assert child.pos_manager.pos == (0, 0)

            self._main_widget = child.get_weakref()
            self.main_widget.set_pos(topleft=(self.padding.left, self.padding.top))
            self._handle_mainwidget_resize()
            self.main_widget.signal.RESIZE.connect(self._handle_mainwidget_resize, owner=None)

    def _handle_mainwidget_resize(self):

        rect = pygame.Rect(self.main_widget.rect)
        rect.left -= self.padding.left
        rect.top -= self.padding.top
        rect.width += self.padding.left + self.padding.right
        rect.height += self.padding.top + self.padding.bottom

        self.max = {
            "x": max(0, rect.width - self.rect.width),
            "y": max(0, rect.height - self.rect.height),
        }

        dx = 0
        if 0 <= (self.rect.width - rect.right) <= - rect.left:
            dx = (self.rect.width - rect.right)
        dy = 0
        if 0 <= (self.rect.height - rect.bottom) <= - rect.top:
            dy = (self.rect.height - rect.bottom)
        if dx or dy:
            self.main_widget.move(dx, dy)

        self.x_scroller.update_from_mainwidget_resize()
        self.y_scroller.update_from_mainwidget_resize()

    def _set_scrollval(self, axis, val):

        if axis == "x":
            self.main_widget.set_pos(left=-val + self.padding.left)
        else:
            self.main_widget.set_pos(top=-val + self.padding.top)

    def handle_mouse_scroll(self, scroll_event):

        old_scroll_time = self._last_scroll_time
        self._last_scroll_time = time.time()
        d = self._last_scroll_time - old_scroll_time

        accelerator = max(20., 1 / d)

        if keyboard.mod.maj:
            scroller = self.x_scroller
        else:
            scroller = self.y_scroller

        if scroll_event.button == 4:
            direction = -1
        else:
            direction = 1

        val = scroller.val + accelerator * direction
        val = min(max(0, val), scroller.maxval)

        scroller.set_val(val)

    def handle_resize(self):

        super().handle_resize()
        self.x_scroller.resize_width(self.rect.w)
        self.y_scroller.resize_height(self.rect.h)
        self._handle_mainwidget_resize()




from baopig import *


class UT_GridLayer_Zone(Zone):

    def __init__(self, *args, **kwargs):
        Zone.__init__(self, *args, **kwargs)

        Layer(self, name="zones_layer", spacing=10)
        z1 = Zone(self, size=("100%", 130), background_color=(150, 150, 150))
        z2 = Zone(self, size=("100%", 130), background_color=(150, 150, 150))
        z3 = Zone(self, size=("100%", 155), background_color=(150, 150, 150))
        z4 = Zone(self, size=("100%", 120), background_color=(150, 150, 150))
        self.default_layer.pack()
        z3.set_style_for(Button, padding=2)

        # Z1
        z1.grid = GridLayer(z1, name="grid_layer", row_height=40, col_width=100)
        assert z1.layers_manager.default_layer == z1.grid
        for row in range(2):
            for col in range(5):
                text = "row:{}\ncol:{}".format(row, col)
                Text(z1, text, row=row, col=col, name=text)

        # Z2
        z2.grid = GridLayer(z2, name="grid_layer")
        for row in range(3):
            for col in range(5):
                text = "row:{}\ncol:{}".format(row, col)
                Text(z2, text, row=row, col=col, name=text)

        class DraggableRectangle(Rectangle, DraggableByMouse):
            def __init__(self, parent, **kwargs):
                Rectangle.__init__(self, parent, **kwargs)
                DraggableByMouse.__init__(self, parent, **kwargs)

        DraggableRectangle(parent=z2, color=(130, 49, 128), size=(30, 30))
        Text(z2, "HI", col=6, row=0)
        Text(z2, "HI", col=7, row=1)
        DraggableRectangle(parent=z2, color=(130, 49, 128), size=(30, 30), col=8, row=2)

        # Z3
        grid = GridLayer(z3, nbrows=5, nbcols=10)

        class RemovableRect(Rectangle, LinkableByMouse):
            def __init__(self, *args, **kwargs):
                Rectangle.__init__(self, *args, **kwargs)
                LinkableByMouse.__init__(self, *args)

            def handle_link(self):
                self.kill()

        import random
        random_color = lambda: [int(random.random() * 255)] * 2 + [128]

        def toggle_col_size(col_index):
            col = grid.get_col(col_index)
            if col.is_adaptable:
                col.set_width(40)
            elif col.get_width() == 40:
                col.set_width(20)
            else:
                col.set_width(None)
            if col.is_adaptable:
                col[-1].kill()

        def toggle_row_size(row_index):
            row = grid.get_row(row_index)
            if row.is_adaptable:
                row.set_height(40)
            elif row.get_height() == 40:
                row.set_height(20)
            else:
                row.set_height(None)
            if row.is_adaptable:
                row[-1].kill()

        def add_rect():
            for row in range(grid.nbrows):
                for col in range(grid.nbcols):
                    if grid._data[row][col] is None:
                        if row is grid.nbrows - 1:
                            Button(z3, "TOG", row=row, col=col, size=(30, 30), catching_errors=True,
                                   command=PrefilledFunction(toggle_col_size, col))
                        elif col is grid.nbcols - 1:
                            Button(z3, "TOG", row=row, col=col, size=(30, 30), catching_errors=True,
                                   command=PrefilledFunction(toggle_row_size, row))
                        else:
                            RemovableRect(z3, color=random_color(), size=(30, 30), col=col, row=row, loc="center")
        Button(z3, "ADD", row=0, col=0, command=add_rect, width=30, height=30)

        def fix():
            if grid.cols_are_adaptable:
                grid.set_row_height(30)
                grid.set_col_width(30)
            else:
                grid.set_row_height(None)
                grid.set_col_width(None)
        Button(z3, "FIX", row=0, col=1, command=fix)

        # Z4
        grid4 = GridLayer(z4, Rectangle, nbrows=4, nbcols=10, row_height=30, col_width=30)
        for row in range(grid4.nbrows):
            for col in range(grid4.nbcols):
                Rectangle(z4, color=(130, 49, 128), row=row, col=col, size=("33%", "33%"), loc="topright")

        buttons_layer = GridLayer(z4, nbrows=3, nbcols=3)

        def click():
            for w in grid4:
                w.set_lock(pos=False)  # TODO : remove ?
                w.pos_manager.config(loc=mouse.hovered_widget.text_widget.text)
                w.set_lock(pos=True)
            grid4.pack()
        Button(z4, row=0, col=0, command=click, text="topleft")
        Button(z4, row=1, col=0, command=click, text="midleft")
        Button(z4, row=2, col=0, command=click, text="bottomleft")
        Button(z4, row=0, col=1, command=click, text="midtop")
        Button(z4, row=1, col=1, command=click, text="center")
        Button(z4, row=2, col=1, command=click, text="midbottom")
        Button(z4, row=0, col=2, command=click, text="topright")
        Button(z4, row=1, col=2, command=click, text="midright")
        Button(z4, row=2, col=2, command=click, text="bottomright")
        buttons_layer.pack(start_pos=(320, 8))

    def load_sections(self):
        self.parent.add_section(
            title="Selector",
            tests=[
                "Any widget from a GridLayer requires 'row' and 'col' attributes",
                "Can't define 'pos' attribute of widgets who will be stored in a grid",
                "Default behavior is to create rows and columns automatically",
                "When the nbrows is set, we can't add a widget who would like to go outside, same for nbcols",
                "A row without defined height adapts to its children, 0 if empty, same for columns",
                "We can set a default size for rows and columns",
                "A widget's hitbox is always inside its cell -> the cell defines the window",
                "We can resize a row without any visual bug inside the row, same for columns -> the window is updated",
                "Resizing a row moves the rows located below, same for columns",
                "Widgets in a grid can't manage their position themselves (non-dragable)",
            ]
        )


# For the PresentationScene import
ut_zone_class = UT_GridLayer_Zone

if __name__ == "__main__":
    from baopig.prefabs.testerscene import TesterScene
    app = Application()
    TesterScene(app, ut_zone_class)
    app.launch()




from baopig import *
from baopig.prefabs.colorchooserdialog import ColorChooserDialog


class UT_ColorChooserDialog_Zone(Zone):
    def __init__(self, *args, **kwargs):
        Zone.__init__(self, *args, **kwargs)

        colorchooser = ColorChooserDialog(self.application)
        Button(self, "Wich color ?", sticky="midtop", pos=(0, 10), command=colorchooser.open)
        text = Text(self, "Color : None", sticky="midtop", pos=(0, 60))
        colorchooser.signal.ANSWERED.connect(lambda color: text.set_text("Color : " + str(color)), owner=None)


# For the PresentationScene import
ut_zone_class = UT_ColorChooserDialog_Zone

if __name__ == "__main__":
    from baopig.prefabs.testerscene import TesterScene
    app = Application()
    TesterScene(app, ut_zone_class).open()
    app.launch()




import pygame
from baopig.lib import Rectangle
from .button import Button


class CheckMark(Rectangle):

    STYLE = Rectangle.STYLE.substyle()
    STYLE.modify(
        height=7,
        width=7,
        color="theme-color-border"
    )

    def __init__(self, checkbox):

        # mid = checkbox.height / 2
        Rectangle.__init__(self, checkbox, sticky="center", ref=checkbox.checkmarkframe)


class CheckMarkFrame(Rectangle):

    STYLE = Rectangle.STYLE.substyle()
    STYLE.modify(
        height=15,
        width=15,
        border_width=2,
        color=(0, 0, 0, 0)
    )

    def __init__(self, checkbox):

        mid = checkbox.rect.height / 2
        Rectangle.__init__(self, checkbox, midleft=(checkbox.padding.left, mid))


class CheckBox(Button):
    # NOTE : connect a function to any state change with the VALIDATE signal

    STYLE = Button.STYLE.substyle()
    STYLE.modify(
        width=100,
        height=35,
        background_color=(0, 0, 0, 0),
        text_style={"font_height": 15, "loc": "midleft", "refloc": "midleft"},
        padding=2,
        spacing=10,
    )
    # TODO : rework checkbox's margin padding and this kind of stuff
    STYLE.create(
        checkmark_class=CheckMark,
        checkmarkframe_class=CheckMarkFrame
    )

    def __init__(self, parent, text=None, is_selected=False, **kwargs):

        Button.__init__(
            self, parent, text=text, **kwargs
        )

        self._is_selected = is_selected

        self._checkmarkframe_ref = CheckMarkFrame(self).get_weakref()
        self._checkmark_ref = CheckMark(self).get_weakref()

        self.text_widget.set_pos(left=self.checkmarkframe.rect.right + self.spacing.left)

        self.checkmark.hide()

        # TODO : all of this, rework with better padding and margin management
        # TODO : when the text is too long, and the width is not specified, the box is lengthen

        # Sliding the text to the right
        assert self.rect.w > self.rect.h
        # self.text_widget.center = ((self.w + self.checkmarkframe.right) / 2, self.h / 2)
        area = pygame.Rect(self.content_rect)
        area.left += self.rect.left
        area.top += self.rect.top
        area_end = area.right
        area.left = self.checkmarkframe.rect.right + self.checkmarkframe.rect.left
        area.w = area_end - area.left
        while self.text_widget.rect.w > area.w:
            if self.text_widget.font.height == 2:
                raise ValueError(f"This text is too long for the text area : {text} (area={area})")
            self.text_widget.font.config(height=self.text_widget.font.height - 1)

    checkmark = property(lambda self: self._checkmark_ref())
    checkmarkframe = property(lambda self: self._checkmarkframe_ref())
    is_selected = property(lambda self: self._is_selected)

    def handle_validate(self, *args, **kwargs):

        self._is_selected = not self.is_selected

        if self.is_selected is True:
            self.checkmark.show()
        else:
            self.checkmark.hide()

        self.command(*args, **kwargs)

        return self.is_selected





from baopig.pybao.objectutilities import *
from baopig.documentation import Container as ContainerDoc
from .imagewidget import Image
from .layer import Layer
from .layersmanager import LayersManager
from .widget_supers import Runable, Widget
from .utilities import *


class BoxRect(pygame.Rect):

    def __init__(self, rect, margin, out=False):

        if out:
            pygame.Rect.__init__(
                self,
                rect.left - margin.left,
                rect.top - margin.top,
                rect.width + margin.left + margin.right,
                rect.height + margin.top + margin.bottom
            )
        else:
            pygame.Rect.__init__(
                self,
                rect.left + margin.left,
                rect.top + margin.top,
                rect.width - margin.left - margin.right,
                rect.height - margin.top - margin.bottom
            )


class ChildrenManager:
    """
    Class for an ordered list of children

    Widgets are sort by overlay, you can access to children sorted by their
    position with children.orderedbypos

    For more efficiency, you can access all the Handler_SceneClose children of a Container
    with Container.children.handlers_sceneclose
    """

    def __init__(self, owner):

        assert isinstance(owner, Container)

        self._owner = owner
        self.all = set()  # accessible via Container.children
        self.containers = set()
        self.handlers_sceneclose = set()
        self.handlers_sceneopen = set()
        self.runables = set()
        self._lists = {
            Handler_SceneClose: self.handlers_sceneclose,
            Handler_SceneOpen: self.handlers_sceneopen,
            Container: self.containers,
            Runable: self.runables,
        }
        self._strong_refs = set()

    def add(self, child):
        """
        This method should only be called by the Widget constructor
        """

        assert child.parent == self._owner
        if child.is_asleep:
            raise PermissionError("A Container cannot contain asleep widgets")

        if child in self.all:
            raise PermissionError(f"{child} already in {self}")

        self.all.add(child)
        self._strong_refs.add(child)
        self._owner.layers_manager.add(child)
        for children_class, children_set in self._lists.items():
            if isinstance(child, children_class):
                children_set.add(child)
        if child.is_visible:
            self._owner._warn_change(child.hitbox)

    def remove(self, child):

        if child.is_asleep:
            raise PermissionError("A Container cannot contain asleep widgets")

        self.all.remove(child)
        self._owner.layers_manager.remove(child)
        for children_class, children_set in self._lists.items():
            if isinstance(child, children_class):
                children_set.remove(child)
        if child.is_visible:
            self._owner._warn_change(child.hitbox)


class Container(ContainerDoc, Widget):
    """
    Abstract class for widgets who need to contain other widgets

    We need the self.container_[action]() functions for recursivity between Container,
    because a container can contain a Handler_SceneOpen without being a Handler_SceneOpen himself

    WARNING : Try to do not override 'container_something' methods
    """

    STYLE = Widget.STYLE.substyle()
    STYLE.create(
        background_color=(0, 0, 0, 0),  # transparent by default
        background_image=None,
        border_color="theme-color-border",
        border_width=0,
        spacing=0,
        padding=0,
    )
    STYLE.set_type("background_color", Color)
    STYLE.set_type("border_color", Color)
    STYLE.set_type("border_width", int)
    STYLE.set_type("spacing", MarginType)
    STYLE.set_type("padding", MarginType)

    # NOTE : if width or height is defined in style, and a background_image is set,
    # the width and height values will be ignored

    def __init__(self, parent, **kwargs):

        if hasattr(self, "_weakref"):  # Container.__init__() has already been called
            return

        self._children_manager = ChildrenManager(self)  # needed in Widget.__init__  TODO : still ?
        self._children_to_paint = WeakSet()  # a set cannot have two same occurences
        self._rect_to_update = None

        Widget.__init__(self, parent, **kwargs)

        # LAYERS - Only layers can guarantie the overlay
        layersmanager_class = LayersManager
        if "layersmanager_class" in kwargs:
            layersmanager_class = kwargs.pop("layersmanager_class")
            assert issubclass(layersmanager_class, LayersManager)
        self.layers_manager = layersmanager_class(self)
        self.layers = self.layers_manager.layers

        # Box attributes
        self._spacing = self.style["spacing"]
        self._border_color = self.style["border_color"]
        self._border_width = self.style["border_width"]
        self._padding = self.style["padding"]
        self._content_rect = BoxRect(self.auto_rect, self.padding)

        # BACKGROUND
        self._background_color = self.style["background_color"]
        self.background_layer = None
        self._background_image_ref = lambda: None
        background_image = self.style["background_image"]
        if background_image is not None:
            self.set_background_image(background_image)
        self.signal.RESIZE.connect(self.handle_resize, owner=None)

        # self._flip()  # TODO : remove flip_all at application launch, because a Zone created
        #                   during the app lifetime is not correctly printed

    children = property(lambda self: self._children_manager.all)
    background_color = property(lambda self: self._background_color)
    default_layer = property(lambda self: self.layers_manager.default_layer)

    # Box attributes
    spacing = property(lambda self: self._spacing)
    # -
    border_width = property(lambda self: self._border_width)
    padding = property(lambda self: self._padding)
    # -
    border_rect = property(lambda self: self.auto_rect)
    content_rect = property(lambda self: self._content_rect)
    # -
    background_image = property(lambda self: self._background_image_ref())
    border_color = property(lambda self: self._border_color)

    def _add_child(self, child):
        self._children_manager.add(child)

    def _container_close(self):

        for cont in self._children_manager.containers:
            cont._container_close()
        for child in tuple(self._children_manager.handlers_sceneclose):  # tuple prevent from in-loop killed widgets
            child.handle_scene_close()

    def _container_open(self):

        for cont in self._children_manager.containers:
            cont._container_open()
        for child in self._children_manager.handlers_sceneopen:
            child.handle_scene_open()

    def _container_paint(self):
        """ Executes paint requests """

        # TODO : solve :     for cont in self._children_manager.containers:
        #                RuntimeError: Set changed size during iteration
        for cont in self._children_manager.containers:
            cont._container_paint()

        if self._children_to_paint:
            for child in tuple(self._children_to_paint):
                if child.is_visible:
                    child.paint()
                    child.signal.NEW_SURFACE.emit()
                    child.send_display_request()
                    if child._dirty == 1:
                        child._dirty = 0
                        self._children_to_paint.remove(child)
                    # LOGGER.debug("Painting {} from container {}".format(child, self))

        rect = self._update_rect()
        if rect:
            self._warn_parent(rect)

    def _container_refresh(self, recursive=False, only_containers=True, with_update=True):

        if recursive:
            for child in self.children:
                if isinstance(child, Container):
                    child._container_refresh(recursive, only_containers, with_update=False)
                elif not only_containers:
                    child.paint()
                    child.signal.NEW_SURFACE.emit()
                    child.send_display_request()
        if with_update:
            self._flip()
        else:
            self._flip_without_update()

    def _container_run(self):

        for cont in self._children_manager.containers:
            cont._container_run()
        for child in self._children_manager.runables:
            if child.is_running:
                child.run()

    def _flip(self):
        """Update all the surface"""

        if self.is_hidden:
            return
        self._flip_without_update()
        self.send_display_request()

    def _flip_without_update(self):
        """Update all the surface, but don't prevent the parent"""

        with paint_lock:  # prevents self._rect_to_update changes during self._update_rect()
            self._rect_to_update = pygame.Rect(self.auto_rect)
            self._update_rect()

    def _remove_child(self, child):
        self._children_manager.remove(child)
        if child in self._children_to_paint:
            self._children_to_paint.remove(child)

    def _update_rect(self):
        """
        How to update a given portion of the application ?

        This method is the answer.
        This container will update by himself the portion to update, storing the result
        into its surface. Then, it asks its parent to update the same portion,
        and its parent will use this container surface.

        But how can a container update its surface ?

        The container create a surface (rect_background) at the rect size. This new surface
        is going to replace a portion of the container surface corresponding to the rect to
        update, once every child have been blited on it.


        surface :         --------- - - - -----------------------

                                        ^
                                        |

        rect_background :            -------
                                     :     :
                                     :     :
                                     :     :
        child3 :        -------------
                                     :     :
        child2 :                ------------- - - - -             <- The solid line is hitbox, the entire line is rect
                                     :     :
        child1 :                ------------------------------
                                     :     :
        background :        --------------------------------------  <- self.surface is filled with background_color
                                     :     :
                                     :     :
                                     :     :
        rect to update :             :-----:

        """

        if self._rect_to_update is None:
            return
        if self.is_hidden:
            return

        with paint_lock:
            rect, self._rect_to_update = self._rect_to_update, None
            self.surface.fill(self.background_color, rect=rect)
            if self._border_width:
                pygame.draw.rect(self.surface, self._border_color, (0, 0) + self.rect.size, self._border_width * 2 - 1)

            for layer in self.layers:
                for child in layer.visible:
                    if child.hitbox.colliderect(rect):
                        try:
                            # collision is relative to self
                            collision = child.hitbox.clip(rect)
                            if collision == child.rect:
                                self.surface.blit(child.surface, child.rect.topleft)
                            else:
                                self.surface.blit(child.surface.subsurface(
                                    (collision.left - child.rect.left, collision.top - child.rect.top)
                                    + collision.size), collision.topleft
                                )
                        except pygame.error:
                            # can be raised from a child.surface who is a subsurface from self.surface
                            assert child.surface.get_parent() is self.surface
                            child._flip_without_update()  # overdraw child.hitbox

            return rect

    def _warn_change(self, rect):
        """Request updates at rects referenced by self"""

        rect = self.auto_hitbox.clip(rect)
        if rect.size == (0, 0):
            return
        try:
            self._rect_to_update.union_ip(rect)
        except AttributeError:
            self._rect_to_update = pygame.Rect(rect)  # from ProtectedHitbox to pygame.Rect

    def _warn_parent(self, rect):
        """Request updates at rects referenced by self"""

        self.send_display_request(rect=(self.rect.left + rect[0], self.rect.top + rect[1]) + tuple(rect[2:]))

    def adapt(self, children_list=None, vertically=True, horizontally=True):
        """
        Resize in order to contain every widget in children_list
        Not supposed to move children_list
        """

        if children_list is None:
            children_list = self.children
        children = tuple(children_list)
        self.resize(
            width=(max(widget.hitbox.right for widget in children) + self.padding.right
                   if children else self.padding.left + self.padding.right) if horizontally else self.rect.w,
            height=(max(widget.hitbox.bottom for widget in children) + self.padding.bottom
                    if children else self.padding.top + self.padding.bottom) if vertically else self.rect.h
        )

    def handle_resize(self):
        """Stuff to do when the container is resized"""

        if self.background_image is not None:
            self.background_image.resize(*self.rect.size)
        self._content_rect = BoxRect(self.auto_rect, self.padding)
        self._flip_without_update()

    def has_layer(self, layer_name):
        return layer_name in (layer.name for layer in self.layers)

    def kill(self):

        self.hide()
        for child in tuple(self.children):
            child.kill()
        super().kill()

    def pack(self, *args, adapt=False, **kwargs):
        for layer in self.layers:
            layer.pack(*args, **kwargs)
        if adapt:
            self.adapt(self.children)

    def set_background_color(self, *args, **kwargs):

        self._background_color = Color(*args, **kwargs)
        self._warn_change(self.auto_hitbox)  # TODO : test it

    def set_background_image(self, surf, background_adapt=True):
        """
        If background_adapt is True, the surf adapts to the zone's size
        Else, the zone's size adapts to the background_image
        """
        if surf is None:
            if self.background_image is not None:
                with paint_lock:
                    self.background_image.kill()
            return
        if background_adapt and surf.get_size() != self.rect.size:
            surf = pygame.transform.scale(surf, self.rect.size)
        if self.background_layer is None:
            self.background_layer = Layer(self, Image, name="background_layer", level=self.layers_manager.BACKGROUND)
        with paint_lock:
            if self.background_image is not None:
                self.background_image.kill()
                assert self.background_image is None
            self._background_image_ref = Image(self, surf, pos=(0, 0), layer=self.background_layer).get_weakref()
            if background_adapt is False:
                self.resize(*self.background_image.size)

    def set_window(self, *args, **kwargs):

        super().set_window(*args, **kwargs)
        self._rect_to_update = pygame.Rect(self.auto_rect)



from baopig.lib.style import Theme

"""

Color values from theme 'default' :

    border = (0, 0, 0)
    content = (0, 100, 125)
    font = (0, 0, 0)
    font_opposite = (255, 255, 255)
    scene_background = (170, 170, 170)
    selection = (167, 213, 255)
    selection_rect = (107, 107, 205)
    
"""


class DarkTheme(Theme):

    def __init__(self):
        super().__init__()
        self.colors.content = (168, 119, 30)
        self.colors.font = (220, 220, 220)
        self.colors.font_opposite = (0, 0, 0)
        self.colors.scene_background = (85, 85, 85)
        self.colors.selection = (24, 0, 200)
        self.colors.selection_rect = (0, 0, 0)


class GreenTheme(Theme):

    def __init__(self):
        super().__init__()
        self.colors.content = (0, 255, 0)
        self.colors.font_opposite = (100, 255, 100)
        self.colors.scene_background = (190, 210, 190)
        self.colors.selection = (0, 255, 0)
        self.colors.selection_rect = (0, 255, 0)

        from baopig.widgets.text import Text
        self.set_style_for(Text, font_file="passeroone", font_height=20)


class PinkyTheme(Theme):

    def __init__(self):
        super().__init__()
        self.colors.border = (61, 12, 51)
        self.colors.content = (255, 117, 163)
        self.colors.scene_background = (255, 214, 220)
        self.colors.selection = (255, 53, 90)
        self.colors.selection_rect = (255, 53, 90)

        from baopig.widgets.text import Text
        self.set_style_for(Text, font_file="inkfree", font_bold=True)


all_themes = {
    "dark": DarkTheme(),
    "default": Theme(),
    "green": GreenTheme(),
    "pinky": PinkyTheme(),
}




import pygame
from .textedit import TextEdit


class LineEdit(TextEdit):  # TODO : mouse's double click at text_widget's right -> select the text
    """
    LineEdit is a TextEdit who only contains 1 line
    """
    STYLE = TextEdit.STYLE.substyle()
    STYLE.modify(
        height=0
    )

    def __init__(self, parent, **kwargs):

        TextEdit.__init__(self, parent, **kwargs)

        assert self.rect.height == 0, "Cannot set a height, LineEdit manages its height itself"

        self.x_scroller.hide()
        self.x_scroller.set_lock(visibility=True)
        self.resize_height(self.padding.top + self.text_widget.rect.height + self.padding.bottom)

        assert self.y_scroller.is_hidden

    def handle_keydown(self, key):

        if key == pygame.K_RETURN:
            self.defocus()
        else:
            super().handle_keydown(key)

    def set_text(self, text):

        if '\n' in text:
            raise PermissionError("A LineEdit cannot contain a backslash")
        else:
            super().set_text(text)
